
# MoorProps
# start of a script to hold the functions that hold the mooring line and anchor property info
# just simple functions for now, can import functions to other scripts if necessary
# very much in progress
# MoorProps started 4-26-21, getLineProps first initialized back in Sept/Oct 2020


import numpy as np
import moorpy as mp
import moorpy.helpers as helpers
import math

"""
def getLineProps(dmm, type="chain", stud="studless", source="Orcaflex-altered", name=""):
    ''' getLineProps version 3.2: Restructuring v3.1 to 'Orcaflex-original' and 'Orcaflex-altered' 
    
    Motivation: The existing public, and NREL-internal references for mooring line component property
    data are either proprietary, or unreliable and incomplete. The best way to derive new equations
    as a function of diameter is to use data from mooring line manufacturer's catalogs. Once developed,
    these new equations will serve as an updated version to compare against existing expressions.
    
    The goal is to have NREL's own library of mooring line property equations, but more research is needed.
    The original Orcaflex equations are the best we have right now and have been altered to include
    a quadratic chain MBL equation rather than a cubic, to avoid negative MBLs.
    Also, different cost models are put in to the altered version to avoid the Equimar data. Many sources
    researched for cost data, coefficients used are close to the old NREL internal data, but still an approximation.
    
    For more info, see the Mooring Component Properties Word doc.

    - This function requires at least one input: the line diameter in millimeters.
    - The rest of the inputs are optional: describe the desired type of line (chain, polyester, wire, etc.),
    the type of chain (studless or studlink), the source of data (Orcaflex-original or altered), or a name identifier
    - The function will output a MoorPy linetype object

    '''
    
    
    
    if source=="Orcaflex-original":
        d = dmm/1000  # orcaflex uses meters https://www.orcina.com/webhelp/OrcaFlex/
        
        if type=="chain":
            c = 1.96e4 #grade 2=1.37e4; grade 3=1.96e4; ORQ=2.11e4; R4=2.74e4
            MBL = c*d**2*(44-80*d)*1000     #[N]  The same for both studless and studlink
            if stud=="studless":
                massden = 19.9*d**2*1000    #[kg/m]
                EA = 0.854e8*d**2*1000      #[N]
                d_vol = 1.8*d               #[m]
            elif stud=="studlink" or stud=="stud":
                massden = 21.9*d**2*1000    #[kg/m]
                EA = 1.010e8*d**2*1000      #[N]
                d_vol = 1.89*d              #[m]
            else:
                raise ValueError("getLineProps error: Choose either studless or stud chain type ")
        
        elif type=="nylon":
            massden = 0.6476*d**2*1000      #[kg/m]
            EA = 1.18e5*d**2*1000           #[N]
            MBL = 139357*d**2*1000          #[N] for wet nylon line, 163950d^2 for dry nylon line
            d_vol = 0.85*d                  #[m]
        elif type=="polyester":
            massden = 0.7978*d**2*1000      #[kg/m]
            EA = 1.09e6*d**2*1000           #[N]
            MBL = 170466*d**2*1000          #[N]
            d_vol = 0.86*d                  #[m]
        elif type=="polypropylene":
            massden = 0.4526*d**2*1000      #[kg/m]
            EA = 1.06e6*d**2*1000           #[N]
            MBL = 105990*d**2*1000          #[N]
            d_vol = 0.80*d                  #[m]
        elif type=="wire-fiber" or type=="fiber":
            massden = 3.6109*d**2*1000      #[kg/m]
            EA = 3.67e7*d**2*1000           #[N]
            MBL = 584175*d**2*1000          #[N]
            d_vol = 0.82*d                  #[m]
        elif type=="wire-wire" or type=="wire" or type=="IWRC":
            massden = 3.9897*d**2*1000      #[kg/m]
            EA = 4.04e7*d**2*1000           #[N]
            MBL = 633358*d**2*1000          #[N]
            d_vol = 0.80*d                  #[m]
        else:
            raise ValueError("getLineProps error: Linetype not valid. Choose from given rope types or chain ")
            
        
        
        # cost
        # Derived from Equimar graph: https://tethys.pnnl.gov/sites/default/files/publications/EquiMar_D7.3.2.pdf
        if type=="chain":
            cost = (0.21*(MBL/9.81/1000))*1.29              #[$/m]
        elif type=="nylon" or type=="polyester" or type=="polypropylene":
            cost = (0.235*(MBL/9.81/1000))*1.29         #[$/m]
        elif type=="wire" or type=="wire-wire" or type=="IWRC" or type=='fiber' or type=='wire-fiber':
            cost = (0.18*(MBL/9.81/1000) + 90)*1.29     #[$/m]
        else:
            raise ValueError("getLineProps error: Linetype not valid. Choose from given rope types or chain ")



    elif source=="Orcaflex-altered":
        d = dmm/1000  # orcaflex uses meters https://www.orcina.com/webhelp/OrcaFlex/
        
        if type=="chain":
            c = 2.74e4 #grade 2=1.37e4; grade 3=1.96e4; ORQ=2.11e4; R4=2.74e4
            MBL = (371360*d**2 + 51382.72*d)*(c/2.11e4)*1000 # this is a fit quadratic term to the cubic MBL equation. No negatives
            if stud=="studless":
                massden = 19.9*d**2*1000        #[kg/m]
                EA = 0.854e8*d**2*1000          #[N]
                d_vol = 1.8*d                   #[m]
            elif stud=="studlink" or stud=="stud":
                massden = 21.9*d**2*1000        #[kg/m]
                EA = 1.010e8*d**2*1000          #[N]
                d_vol = 1.89*d                  #[m]
            else:
                raise ValueError("getLineProps error: Choose either studless or stud chain type ")
                
            #cost = 2.5*massden   # a ballpark for R4 chain
            #cost = (0.58*MBL/1000/9.81) - 87.6          # [$/m] from old NREL-internal
            #cost = 3.0*massden     # rough value similar to old NREL-internal
            cost = 2.585*massden     # [($/kg)*(kg/m)=($/m)]
            #cost = 0.0
        
        elif type=="nylon":
            massden = 0.6476*d**2*1000      #[kg/m]
            EA = 1.18e5*d**2*1000           #[N]
            MBL = 139357*d**2*1000          #[N] for wet nylon line, 163950d^2 for dry nylon line
            d_vol = 0.85*d                  #[m]
            cost = (0.42059603*MBL/1000/9.81) + 109.5   # [$/m] from old NREL-internal
        elif type=="polyester":
            massden = 0.7978*d**2*1000      #[kg/m]
            EA = 1.09e6*d**2*1000           #[N]
            MBL = 170466*d**2*1000          #[N]
            d_vol = 0.86*d                  #[m]
            
            #cost = (0.42059603*MBL/1000/9.81) + 109.5   # [$/m] from old NREL-internal
            #cost = 1.1e-4*MBL               # rough value similar to old NREL-internal
            cost = 0.162*(MBL/9.81/1000)    #[$/m]
            
        elif type=="polypropylene":
            massden = 0.4526*d**2*1000      #[kg/m]
            EA = 1.06e6*d**2*1000           #[N]
            MBL = 105990*d**2*1000          #[N]
            d_vol = 0.80*d                  #[m]
            cost = 1.0*((0.42059603*MBL/1000/9.81) + 109.5)   # [$/m] from old NREL-internal
        elif type=="hmpe":
            massden = 0.4526*d**2*1000      #[kg/m]
            EA = 38.17e6*d**2*1000          #[N]
            MBL = 619000*d**2*1000          #[N]
            d_vol = 1.01*d                  #[m]
            cost = (0.01*MBL/1000/9.81)     # [$/m] from old NREL-internal
        elif type=="wire-fiber" or type=="fiber":
            massden = 3.6109*d**2*1000      #[kg/m]
            EA = 3.67e7*d**2*1000           #[N]
            MBL = 584175*d**2*1000          #[N]
            d_vol = 0.82*d                  #[m]
            cost = 0.53676471*MBL/1000/9.81             # [$/m] from old NREL-internal
        elif type=="wire-wire" or type=="wire" or type=="IWRC":
            massden = 3.9897*d**2*1000      #[kg/m]
            EA = 4.04e7*d**2*1000           #[N]
            MBL = 633358*d**2*1000          #[N]
            d_vol = 0.80*d                  #[m]
            #cost = MBL * 900./15.0e6
            #cost = (0.33*MBL/1000/9.81) + 139.5         # [$/m] from old NREL-internal
            cost = 5.6e-5*MBL               # rough value similar to old NREL-internal
        else:
            raise ValueError("getLineProps error: Linetype not valid. Choose from given rope types or chain ")


    elif source=="NREL":
        '''
        getLineProps v3.1 used to have old NREL-internal equations here as a placeholder, but they were not trustworthy.
         - The chain equations used data from Vicinay which matched OrcaFlex data. The wire rope equations matched OrcaFlex well,
           the synthetic rope equations did not
           
        The idea is to have NREL's own library of mooring line property equations, but more research needs to be done.
        The 'OrcaFlex-altered' source version is a start and can change name to 'NREL' in the future, but it is
        still 'OrcaFlex-altered' because most of the equations are from OrcaFlex, which is the best we have right now.
        
        Future equations need to be optimization proof = no negative numbers anywhere (need to write an interpolation function)
        Can add new line types as well, such as Aramid or HMPE
        '''
        pass
        
    
    
    
    # Set up a main identifier for the linetype. Useful for things like "chain_bot" or "chain_top"
    if name=="":
        typestring = f"{type}{dmm:.0f}"
    else:
        typestring = name
    
    
    notes = f"made with getLineProps - source: {source}"
    

    return mp.LineType(typestring, d_vol, massden, EA, MBL=MBL, cost=cost, notes=notes, input_type=type, input_d=dmm)
"""

#----NOTES----

#def getAnchorProps(value, anchor="drag-embedment", soil_type='medium clay',uhc_mode = True, method = 'static', display=0):
    # need to make 'value' input all the same as either Mass or UHC
    #'value' can be both capacity or mass, 'uhc_mode' defines what it is. However, this method does not take into account fx, fz loading directions or safety factor

# applied safety factors make input UHC smaller resulting in smaller mass > uhc_mode = True, calulating for capacity | INPUT: mass | does not account for this 
    #added 'capacity_sf' variable: outputs capacity for applied safety factor, while showing original input capacities (fx,fz)

# VLA calculation for thickness may be wrong: may not need to divide by 4. When we dont it matched STEVMANTA table (in excel sheet)

# Do i want to scale fx,fz by 20% when considering dynamic loads like in Moorpy 'capacity_x = 1,2*fx ; capacity_z = 1.2*fz

# Need to validate suction curves and compare to Ahmed's curves. His is pure horizontal while these ones are for both horizontal and vertical 

#--------------

def getAnchorMass( uhc_mode = True, fx=0, fz=0, mass_int=0, anchor="drag-embedment", soil_type='medium clay', method = 'static', aprops = None, display=0):
    '''Calculates anchor required capacity, mass, and cost based on specified loadings and anchor type
    
    Parameters
    ----------
    uhc_mode: boolean
        True: INPUT Mass to find UHC
        False: INPUT fx,fz to find Mass and UHC
    fx : float
        horizontal maximum anchor load [N]
    fz : float 
        vertical maximum anchor load [N]
    massint : float 
        anchor mass [kg]    
    anchor : string
        anchor type name
    soil_type: string
        soil type name
    method: string
        anchor analysis method for applied safety factors 

    Returns
    -------
    UHC: float
        required anchor ultimate holding capacity [N]
    mass: float
        anchor mass [kg]
    info: dict
        dictionary
    '''

    # # --- Load PointProps ---
    # if not aprops:
    #     props = helpers.loadPointProps(None) 
    #     # load Anchor props
    #     aprops = props["AnchorProps"]
    #     # In progress. Eventually these hardcoded coefficients should be brought over to PointProps_default.yaml


    #--MAGIC NUMBERS -- assign varibles to values
    density_st = 7.85*1000       #[kg/m^3]
    gravity = 9.81               #[m/sec^2]? or [N/kg], if so may need to add conversion *1000 so [kN/mT]
    info = {}


    # Define anchor type
    if anchor == "drag-embedment":
       
        # Define soil type (coefficients are from Vryhof for Stevpris MK6)
        if soil_type == 'soft clay':
            a, b = 509.96, 0.93
        elif soil_type == 'medium clay':
            a, b = 701.49, 0.93
        elif soil_type == 'hard clay' or soil_type == 'sand':
            a, b = 904.21, 0.92
        else:
            raise ValueError('Error: Invalid soil type')
        
        # Define mode
        # Calculate capacity | INPUT: Mass in kg 
        if uhc_mode:
            uhc = a * (mass_int/1000)**b*1000           #[N] (from Vryhof eqn)
            # y = m*(x)^b

            print(f"Mass input: {mass_int} -- UHC: {uhc}")
            info["Mass"] = mass_int   #[kg]
            info["UHC"] = uhc               #[N]
            
        
        # Calculate mass | INPUT: UHC 
        else:
            # Define method for safety factor
            if method == 'static':
                uhc = 1.8*fx     #[N]
            elif method == 'dynamic':
                uhc = 1.5*fx      #[N]
            else:
                raise Exception("Error - invalid method")

            mass = (uhc /1000 /a)**(1/b)*1000     # [kg] (from Vryhof eqn)
            
            print(f"UHC input: {fx} -- Mass: {mass}")
            info["UHC"] = uhc           #[N]
            info["Mass"] = mass                 #[kg]
        
    elif anchor == "VLA":
        
        # Vryhof coefficients
        if soil_type == 'soft clay':
            c, d = 0.003, -0.2
        elif soil_type == 'medium clay':
            c, d = 0.0017, -0.32
        else:
            raise ValueError('Error: Invalid soil type')

        if uhc_mode: # Calculate capacity | INPUT: Mass
            #t2_m_ratio = 20933.3                            #[m^2]

            #Ericka note - I don't know where this ratio is coming from. Matt may have estimated from Vryhof manual
            t2_m_ratio = 1308.33
            
            #Ericka note -not sure about this equation - should it be to the 1/2 power? 
            thickness = (mass_int/1000 / t2_m_ratio)**(1/3)      #[m]     #DONE: magic number - assign to variable (think its from t^2/A = 6000; thickness ratio)
            area = mass_int/ (thickness * density_st)            #[m^2]   Area (1-30)
            uhc = (area - d) / c * 1000                          #[N] Vryhof equation


            print(f"Mass input: {mass_int} -- UHC: {uhc}, Area: {area}")
            info["Mass"] = mass_int                 #[kg]
            info["UHC"] = uhc                       #[N]
            info["Area"] = area                     #[m^2]

        else: # Calculate mass | INPUT: UHC
            capacity_x = 2.0 * fx  # N
            capacity_z = 2.0 * fz  # N
            uhc = np.linalg.norm([capacity_x,capacity_z])
            
            #Vryhof equation
            area = c * uhc/1000 + d          #[m^2] Vryhof equation
            # y = m*x + b
            
            #Ericka note - I don't know where this ratio is coming from. Matt may have estimated from Vryhof manual
            t2_a_ratio = 0.006                  #[m^2]
            if area < 0:
                thickness = 0
                raise ValueError ("Error: Negative area)")
            else:
                #thickness = math.sqrt(t2_a_ratio * area)/4          #[m]    #### why divide by 4 - looking at this in exel sheet
                thickness = math.sqrt(t2_a_ratio * area)
                mass = area * thickness * density_st                #[kg]


                print(f"UHC input: fx:{fx} fz:{fz} -- Mass: {mass}, Area: {area}")
                info["UHC"] = uhc   #[N]
                info["Mass"] = mass                 #[kg]
                info["Area"] = area                 #[m^2]

    
    ###### VERIFICATION REQUIRED, THESE CURVES CONSIDER INCLINED LOADS -> While Ahmed's is pure horizontal 
    elif anchor == "suction":
        
        # c and d numbers are from ABS FOWT Anchor Report
        # m and b numbers are from Brian's excel sheet curve fit to Mass vs UHC graph
        if soil_type == 'soft clay':
            m = 88.152
            b = 1.1058
            c_L, d_L = 1.1161, 0.3442
            c_D, d_D = 0.3095, 0.2798
            c_T, d_T = 2.058, 0.2803
            
        elif soil_type == 'medium clay':
            m = 384.15
            b = 0.8995
            c_L, d_L = 0.5166, 0.3995
            c_D, d_D = 0.126, 0.3561
            c_T, d_T = 0.8398, 0.3561
           
        else:
            raise ValueError("Invalid Soil Type")
        
        
        if uhc_mode: # Calculate capacity | INPUT: Mass
           
            ### ABS (Matts curves) slightly off  (from Brian's excel sheet curve fit for Mass vs UHC)
            uhc = m * (mass_int/1000) ** b * 1000 # [N]


            print(f"Mass INPUT: {mass_int} -- UHC: {uhc}")
            info["Mass"] = mass_int       #[kg]
            info["UHC"] = uhc                   #[N]

        else: # Calculate mass | INPUT: UHC
        
            capacity_x = 1.6*fx                           #[N]
            #capacity_x = fx
            capacity_z = 2.0*fz                           #[N]
            uhc = np.linalg.norm([capacity_x, capacity_z])  #[kN]

            # Equations from ABS Anchors Report
            L = c_L * (uhc/1000) **d_L                 #[m]
            D = c_D * (uhc/1000) **d_D                 #[m]
            T = (c_T * (uhc/1000) **d_T) /1000        #[m] 

            # y = m*(x)^b
            
            # volume of one end + open cylinder wall
            volume = (math.pi/4 * D ** 2 + math.pi * D * L) * T     #[m^3]
            mass = volume * density_st                              #[kg]


            print(f"UHC input: fx:{fx} fz:{fz} -- Mass: {mass}")
            info["UHC"] = uhc           #[N]
            info["Mass"] = mass                 #[kg]
            #info["Length"] = L
          
    elif anchor == "gravity":

        capacity_x = 1.6*fx                           #[N]
        capacity_z = 2.0*fz                           #[N]
        uhc = np.linalg.norm([capacity_x, capacity_z])  
 
        mass = max(capacity_x / gravity, capacity_z / gravity) # 2x safety factor for vertial 1.6x for horizontal per API standards. Use the larger value. Max load in kg * FOS = mass
        # Alterative would be to use 4x the horizontal load per WEC mooring advice from UW APL. 

        info["UHC"] = uhc           #[N]
        info["Mass"] = mass                 #[kg]

    #-----IN PROGRESS -----------------------------------

    elif anchor == "SEPLA":
        # Ericka note - we think these coefficients come from Ahmed's intermediate model
        if soil_type == 'soft clay':
            c_B, d_B = 0.0225, 0.5588
            c_L, d_L = 0.0450, 0.5588
            c_T, d_T = 0.0006, 0.5588
            
        else:
            raise ValueError("Invalid Soil Type")
        
        if uhc_mode: # Calculate UHC | INPUT: Mass       ### A little bit off from uhc = FALSE 

            # from excel
            m = 1557.4
            b = 0.5956
            uhc = m *(mass_int/1000) ** b  *10000 # [N]

        
            print(f'Work in progress -- Mass input: {mass_int} -- UHC {uhc}')
            info["Mass"] = mass_int
            info["UHC"] = uhc

        else: # Calculate Mass | INPUT: UHC
            capacity_x = 2.0*fx
            uhc = capacity_x
        
            B = c_B * (uhc/1000) ** d_B        #[m]
            L = c_L * (uhc/1000) ** d_L        #[m]
            T = c_T * (uhc/1000) ** d_T        #[m]
            # y = m*(x)^b

            area = B*L                          #[m^2]
            volume = area * T                   #[m^3]
            mass = volume * density_st          #[mT]


            print(f"UHC input: fx:{fx} fz:{fz} -- Mass {mass}")
            info["UHC"] = uhc   #[N]
            info["Mass"] = mass                 #[kg]
            info["Area"] = area                 #[m^2]

    elif anchor == "micropile":
        raise ValueError ("Not supported yet")    


    else:
        return Exception("Error - invalid anchor type")
    


    if uhc_mode:
        mass = mass_int

    return uhc, mass, info
    #return info


def getAnchorCostOld(fx, fz, type="drag-embedment",soil_type='medium clay', method = 'static'):
    ''' applies factors to material cost '''
    

    uhc, mass, info = getAnchorMass( uhc_mode = False, fx=fx, fz=fz, anchor= type, soil_type=soil_type, method = method)       
    euros2dollars = 1.18    # the number of dollars there currently are in a euro (3-31-21)
    
    if type == "drag-embedment":

 
        anchorMatCost = 6.7*mass # $ per kg mass
        anchorInstCost = 163548*euros2dollars       # installation cost
        anchorDecomCost = 228967*euros2dollars      # decommissioning cost
        
    elif type == "suction":

        anchorMatCost = 10.25 *mass # $ per kg mass
        anchorInstCost = 179331*euros2dollars   # installation cost
        anchorDecomCost = 125532*euros2dollars  # decommissioning cost
    
        
    elif type == "micropile":   
        
        raise ValueError('Micropile material costs are not yet supported')
        #anchorMatCost = 0.48 * mass # $ per kg mass
        anchorInstCost = 0                                  # installation cost
        anchorDecomCost = 0                                 # decommissioning cost
    
    elif type == "plate":   # cross between suction and plate
        
        raise ValueError('Plate material costs are not yet supported')
        #anchorMatCost = 0.45 * mass# $ per kg mass
        anchorInstCost = 0                      # installation cost
        anchorDecomCost = 0                     # decommissioning cost
    
    else:
        raise ValueError(f'getAnchorProps received an unsupported anchor type ({type})')
        
    # mooring line sizing:  Tension limit for QS: 50% MBS.  Or FOS = 2
    

    return anchorMatCost, anchorInstCost, anchorDecomCost, info   # [USD]

def getAnchorCost(fx = None, fz = None, type="drag-embedment", mass = None, area = None, soil_type='medium clay', method = 'static', aprops = None):
    '''Calculates anchor cost, and calls getAnchorMass if mass is not provided and design forces are. 
    Costs based on the WEC mooring cost modeling work. 
    
    Parameters
    ----------
    fx : float
        horizontal maximum anchor load [N]
    fz : float 
        vertical maximum anchor load [N] 
    type : string
        anchor type name
    mass : float 
        anchor mass [kg]   
    area : float 
        anchor plate area (for VLA) [m^2]  
    soil_type: string
        soil type name
    method: string
        anchor analysis method for applied safety factors 
    aprops: dict
        an AnchorProps structure containing coefficient data. This can by found in PointProps as PointProps["AnchorProps"]

    Returns
    -------
    anchorMatCost : float
        the anchor material cost [2024$]
    anchorInstCost : float
        the anchor installation cost [$]
    anchorDecomCost : float
        the anchor decomissioning cost [$]
    info: dict
        dictionary
    '''
    
    # >>> This function can be moved into Point <<<

    # --- Load PointProps ---
    if not aprops:
        props = helpers.loadPointProps(None)
        # load Anchor props
        aprops = props["AnchorProps"]

    # --- Get Anchor Size if not provided ---
    if (mass == None and type != "VLA") or (area == None and type == "VLA"):
        if fx == None or fz == None:
            raise ValueError(f"Horizontal and vertical loads required for size calculation of anchor type: {type}")
        uhc, mass, info = getAnchorMass( uhc_mode = False, fx=fx, fz=fz, anchor = type, soil_type=soil_type, method = method, aprops = aprops)
        area = info.get("Area", 0.0) # if area is not calculated, make zero
    else:
        info = dict(UHC = "", Mass = mass, Area = area, Notes = "")   
    
    # calculate cost
    if type in aprops.keys(): # some checking for if value is not avaible (coud be give a none value as deafult if not proivded and check for error)
        anchorMatCost = aprops[type]["matcost"] + aprops[type]["matcost_m"] * mass + aprops[type]["matcost_m2"] * mass**2 + aprops[type]["matcost_m3"] * mass**3 + aprops[type]["matcost_a"] * area + aprops[type]["matcost_a2"] * area**2 + aprops[type]["matcost_a3"] * area**3
        anchorInstCost = aprops[type]["instcost"] + aprops[type]["instcost_m"] * mass + aprops[type]["instcost_m2"] * mass**2 + aprops[type]["instcost_m3"] * mass**3 + aprops[type]["instcost_a"] * area + aprops[type]["instcost_a2"] * area**2 + aprops[type]["instcost_a3"] * area**3
        anchorDecomCost = aprops[type]["decomcost"] + aprops[type]["decomcost_m"] * mass + aprops[type]["decomcost_m2"] * mass**2 + aprops[type]["decomcost_m3"] * mass**3 + aprops[type]["decomcost_a"] * area + aprops[type]["decomcost_a2"] * area**2 + aprops[type]["decomcost_a3"] * area**3
        if anchorMatCost == 0.0 or anchorInstCost == 0.0 or anchorDecomCost == 0.0:
            info["Notes"] = f"A cost of zero indicates respective cost curve not provided for the anchor type {type}"
    else:
        raise ValueError(f'The anchor type {type} was not found in the AnchorProps library')

    return anchorMatCost, anchorInstCost, anchorDecomCost, info   # [USD]


"""
def getAnchorProps(fx, fz, type="drag-embedment", display=0):
    ''' ****OLD VERSION**** Calculates anchor required capacity and cost based on specified loadings and anchor type'''
    
    # for now this is based on API RP-2SK guidance for static analysis of permanent mooring systems
    # fx and fz are horizontal and vertical load components assumed to come from a dynamic (or equivalent) analysis.
    
    # mooring line tenLimit specified in yaml and inversed for a SF in constraints
    # take the line forces on the anchor and give 20% consideration for dynamic loads on lines
    # coefficients in front of fx and fz in each anchorType are the SF for that anchor for quasi-static (pages 30-31 of RP-2SK)
    
    # scale QS loads by 20% to approximate dynamic loads
    # fx = 1.2*fx
    # fz = 1.2*fz
    
    # note: capacity is measured here in kg force
    
    euros2dollars = 1.18    # the number of dollars there currently are in a euro (3-31-21)
    
    if type == "drag-embedment":
        capacity_x = 1.5*fx/9.81
        
        fzCost = 0
        #if fz > 0: 
        #    fzCost = 1e3*fz
        #    if display > 0:  print('WARNING: Nonzero vertical load specified for a drag embedment anchor.')   
        
        anchorMatCost = 0.188*capacity_x + fzCost   # material cost
        anchorInstCost = 163548*euros2dollars       # installation cost
        anchorDecomCost = 228967*euros2dollars      # decommissioning cost
        
    elif type == "suction":
        capacity_x = 1.6*fx/9.81
        capacity_z = 2.0*fz/9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])    # overall capacity, assuming in any direction for now
        anchorMatCost = 1.08*capacity           # material cost
        anchorInstCost = 179331*euros2dollars   # installation cost
        anchorDecomCost = 125532*euros2dollars  # decommissioning cost
    
    elif type == 'deadweight-granite':
        capacity_x = 1.6*fx/9.81    # no safety factors given explicitly for deadweight anchors in the standards; assuming these safety factors are the same as piles/gravity/plates
        capacity_z = 2.0*fz/9.81    # no safety factors given explicitly for deadweight anchors in the standards; assuming these safety factors are the same as piles/gravity/plates
        capacity = np.linalg.norm([capacity_x, capacity_z])
        anchorMatCost = 0.05*capacity   # cost of a granite deadweight anchor is about $50 per ton (from Senu)
        anchorInstCost = 0
        anchorDecomCost = 0
    
    elif type == 'deadweight-concrete':
        capacity_x = 1.6*fx/9.81    # no safety factors given explicitly for deadweight anchors in the standards; assuming these safety factors are the same as piles/gravity/plates
        capacity_z = 2.0*fz/9.81    # no safety factors given explicitly for deadweight anchors in the standards; assuming these safety factors are the same as piles/gravity/plates
        capacity = np.linalg.norm([capacity_x, capacity_z])
        anchorMatCost = 0.075*capacity   # cost of a concrete deadweight anchor is about $75 per ton (from Senu)
        # reinforced concrete is about $1000 per ton ($1 per kg), but we don't need reinforced concrete for deadweight anchors
        anchorInstCost = 0
        anchorDecomCost = 0
    
    elif type == "plate":
        capacity_x = 2.0*fx/9.81
        capacity_z = 2.0*fz/9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])    # overall capacity, assuming in any direction for now
        raise ValueError("plate anchors not yet supported")
        
    elif type == "micropile":    # note: no API guidance on this one
        capacity_x = 2.0*fx/9.81
        capacity_z = 2.0*fz/9.81
        capacity = np.linalg.norm([capacity_x, capacity_z]) # overall capacity, assuming in any direction for now
        anchorMatCost = (200000*1.2/500000)*capacity        # [(Euros*($/Euros)/kg)*kg] linear interpolation of material cost
        anchorInstCost = 0                                  # installation cost
        anchorDecomCost = 0                                 # decommissioning cost
    
    elif type == "SEPLA":   # cross between suction and plate
        capacity_x = 2.0*fx/9.81
        capacity_z = 2.0*fz/9.81
        capacity = np.linalg.norm([capacity_x, capacity_z])    # overall capacity, assuming in any direction for now
        anchorMatCost = 0.45*capacity           # material cost
        anchorInstCost = 0                      # installation cost
        anchorDecomCost = 0                     # decommissioning cost
    
    else:
        raise ValueError(f'getAnchorProps received an unsupported anchor type ({type})')
        
    # mooring line sizing:  Tension limit for QS: 50% MBS.  Or FOS = 2
    

    return anchorMatCost, anchorInstCost, anchorDecomCost   # [USD]
"""



import numpy as np
import yaml

from moorpy.system import System
from moorpy.body import Body
from moorpy.point import Point
from moorpy.line import Line, from2Dto3Drotated
from moorpy.lineType import LineType
from moorpy.helpers import (rotationMatrix, rotatePosition, getH, printVec, 
                            set_axes_equal, dsolve2, SolveError, MoorPyError, 
                            loadLineProps, getLineProps, read_mooring_file, 
                            printMat, printVec, getInterpNums, unitVector,
                            getFromDict, addToDict)



class Subsystem(System, Line):
    '''A class for a mooring line or dynamic cable subsystem.
    It includes Line sections but also can fit into a larger System
    the same way a Line can.
    
    A subsystem tracks its own objects in its local coordinate system.
    This local coordinate system puts the x axis along the line heading,
    from anchor to fairlead, with the anchor at x=0.
    
    For a multi-section line or cable (the main reason to use a SubSystem),
    the free DOFs of the points are contained in the SubSystem and are not 
    seen by the parent System. Solving their equilibrium is an internal
    solve nested within the larger System's equilibrium solve.
    
    The first and last points in the subsystem are set as coupled so that
    their stiffness matrices can be computed and passed back the same way as
    those of a line's two ends.
    
    Line types should be referenced from the overall System... [how?]
    
    Key adjustments: replace staticSolve from Line to do an internal 
    equilibrium solve using System.solveEquilibrium.
    '''
    
    
    def __init__(self, mooringSys=None, num=0, depth=0, rho=1025, g=9.81, 
                 lineProps=None, **kwargs):
        '''Shortened initializer for just the SubSystem aspects.
        
        How should this be initialized???
        '''
        
        # some basics for compatibility with Line objects when used in a System lineList
        self.sys = mooringSys  # store a reference to the overall mooring system (instance of System class)
        self.number = num
        self.isRod = False
        self.type = {}  # because Line objects have a type dict...
        
        # lists to hold mooring system objects
        self.bodyList = []
        self.rodList = []
        self.pointList = []
        self.lineList = []
        self.lineTypes = {}  # dict indexed as a list [0:n-1] corresponding to lineList. Contains *references* to lineType dicts.
        self.rodTypes = {}
        
        # some basic subsystem things
        self.rA = np.zeros(3)
        self.rB = np.zeros(3)
        self.LH = 0
        self.th = 0
        self.sin_th = 0
        self.cos_th = 1
        
        # load mooring line property scaling coefficients for easy use when creating line types
        self.lineProps = loadLineProps(lineProps)
        
        # the ground body (number 0, type 1[fixed]) never moves but is the parent of all anchored things
        self.groundBody = Body(self, 0, 1, np.zeros(6))
        
        # constants used in the analysis
        self.depth = depth  # water depth [m]
        self.rho   = rho    # water density [kg/m^3]
        self.g     = g      # gravitational acceleration [m/s^2]
        
        # Set position tolerance to use in equilibrium solves [m]
        self.eqtol = getFromDict(kwargs, 'eqtol', default=0.01)
        
        # water current - currentMod 0 = no current; 1 = steady uniform current
        self.currentMod = 0         # flag for current model to use
        self.current = np.zeros(3)  # current velocity vector [m/s]
        if 'current' in kwargs:
            self.currentMod = 1
            self.current = getFromDict(kwargs, 'current', shape=3)
            
        # flag to indicate shared line or suspended cable being modeled as symmetric
        self.shared = getFromDict(kwargs, 'shared', dtype=bool, default=False)
        
        self.span   = getFromDict(kwargs, 'span', default=0)  # horizontal end-end distance [m]
        self.rad_fair = getFromDict(kwargs, 'rad_fair', default=0)  # [m] fairlead radius [m]
        self.z_fair   = getFromDict(kwargs, 'z_fair'  , default=0)  # [m] fairlead z coord [m]
        
        # the old rAFair input for the position of a second fairlead of a shaerd mooring line is gone
        # currently we assume rad_fair and z_fair are same for both ends of a shared line...
        
        self.offset = 0  # used for debugging
        
        # seabed bathymetry - seabedMod 0 = flat; 1 = uniform slope, 2 = grid
        self.seabedMod = 0
        
        if 'xSlope' in kwargs or 'ySlope' in kwargs:
            self.seabedMod = 1
            self.xSlope = getFromDict(kwargs, 'xSlope', default=0)
            self.ySlope = getFromDict(kwargs, 'ySlope', default=0)
        
        #if 'bathymetry' in kwargs:
            #self.seabedMod = 2
            #self.bathGrid_Xs, self.bathGrid_Ys, self.bathGrid = self.readBathymetryFile(kwargs['bathymetry'])
        '''
        if 'bathymetry' in kwargs:
            self.seabedMod = 2
            if isinstance(kwargs['bathymetry'],str):
                self.bathGrid_Xs, self.bathGrid_Ys, self.bathGrid = readBathymetryFile(kwargs['bathymetry'])
            elif isinstance(kwargs['bathymetry'],dict):
                bath_dictionary = kwargs['bathymetry']
                # grid sent in, just assign to properties
                self.bathGrid_Xs = bath_dictionary['x']
                self.bathGrid_Ys = bath_dictionary['y']
                self.bathGrid = bath_dictionary['depth'] 
        '''
        # Note, System and Subsystem bathymetry can be set after creation 
        # by setBathymetry, which uses link to existing data, for efficiency.
        
        # initializing variables and lists        
        self.nDOF = 0       # number of (free) degrees of freedom of the mooring system (needs to be set elsewhere)        
        self.freeDOFs = []  # array of the values of the free DOFs of the system at different instants (2D list)
        
        self.nCpldDOF = 0   # number of (coupled) degrees of freedom of the mooring system (needs to be set elsewhere)        
        self.cpldDOFs = []  # array of the values of the coupled DOFs of the system at different instants (2D list)
        
        self.display = 0    # a flag that controls how much printing occurs in methods within the System (Set manually. Values > 0 cause increasing output.)
        
        self.MDoptions = {} # dictionary that can hold any MoorDyn options read in from an input file, so they can be saved in a new MD file if need be
        self.qs = 1         # flag that it's a MoorPy analysis, so System methods don't complain. One day should replace this <<<
    
        self.dynamic_stiffness_activated = False  # flag turned on when dynamic EA values are activate
    
    
    def updateBathymetry(self):
        '''Updates the bathymetry description underneath the Subsystem based on
        global bathymetry data stored in the parent System if it exists.
        Bathymetry under the subsystem is represented as an x-depth curve with
        node points corresponding to wherever the subsystem cross over an edge
        of the bathymetry grid.
        '''
        
        dx = self.rB[0] - self.rA[0]
        dy = self.rB[1] - self.rA[1]
        
        LH = np.hypot(dx, dy)  # horizontal distance between ends
        
        if self.sys and hasattr(self.sys, 'bathGrid_Xs'):  # only do this if there is a parent System AND bathymetry grid
            
            xs_loc = []  # local x value along subsystem from end A
            depths = []
            
            # Find all places where the subsystem crosses the bathGrid_Xs
            if abs(dx) > 0:
                for x in self.sys.bathGrid_Xs:  # look along each grid line
                    # Check if x is between rA and rB:
                    if (self.rA[0] < x and x < self.rB[0]) or (self.rA[0] > x and x > self.rB[0]):
                        
                        y = self.rA[1] + (x - self.rA[0])*dy/dx  # corresponding y value
                        
                        depth, _ = self.sys.getDepthFromBathymetry(x, y) # depth
                        
                        xs_loc.append( (x - self.rA[0])*LH/dx )  # hotizontal length along sybsystem
                        depths.append(depth)
        
        
            # Find all places where the subsystem crosses the bathGrid_Ys
            if abs(dy) > 0:
                for y in self.sys.bathGrid_Ys:  # look along each grid line
                    # Check if x is between rA and rB:
                    if (self.rA[1] < y and y < self.rB[1]) or (self.rA[1] > y and y > self.rB[1]):
                        
                        x = self.rA[1] + (y - self.rA[1])*dx/dy  # corresponding y value
                        
                        depth, _ = self.sys.getDepthFromBathymetry(x, y) # depth
                        
                        xs_loc.append( (y - self.rA[1])*LH/dy )  # hotizontal length along sybsystem
                        depths.append(depth)
        
            # Sort xs_loc and depths by increasing xs_loc values
            xs_loc = np.array(xs_loc)
            depths = np.array(depths)
            idx = np.argsort(xs_loc)
            xs_loc_sorted = xs_loc[idx]
            depths_sorted = depths[idx]
            # >>> TO DO <<<
            
            # Store the updated 1D representation in the Subsystem
            self.bathGrid_Xs = np.array(xs_loc_sorted)
            self.bathGrid_Ys = np.array([0])
            self.bathGrid    = np.array([depths_sorted])
            
        else:
            pass  # MH: can't think of anything to do for isolated subsystems at the moment...
        
    
    def initialize(self, daf_dict={}):
        '''Initialize the Subsystem including DAFs.'''
        
        # Count number of line sections and total number of nodes
        self.nLines = len(self.lineList)
        self.nNodes = np.sum([l.nNodes for l in self.lineList]) - self.nLines + 1
        
        # Use the System initialize method
        System.initialize(self)
        
        # dynamic amplication factor for each line section, and anchor forces 
        # (DAFS[-2] is for vertical load, DAFS[-1] is for horizontal load)
        self.DAFs = getFromDict(daf_dict, 'DAFs', shape=self.nLines+2, default=1.0)  
        
        # mean tension [N] of each line section end [section #, end A/B] for any given mean offset
        self.Te0 = np.zeros([self.nLines,2])  # undispalced values
        self.TeM = np.zeros([self.nLines,2])  # mean offset values
        self.TeD = np.zeros([self.nLines,2])  # dynamic values
        self.TeDmin = np.zeros([self.nLines,2])  # dynamic minimum values
        
        # adjustment on laylength... positive means that the dynamic lay length is greater than linedesign laylength 
        self.LayLen_adj = getFromDict(daf_dict, 'LayLen_adj', shape=0, default=0.0) 
        
        # Fatigue damage over all nodes (placeholder for now)
        self.damage = np.zeros(self.nNodes)
        
    
    def makeGeneric(self, lengths, types, connectors=[], suspended=0, nSegs=40):

        '''Creates a cable of n components going between an anchor point and
        a floating body (or a bridle point). If shared, it goes between two
        floating bodies.

        Parameters
        ----------
        lengths : list of floats
            List of each section's length. This also implies the number of
            sections.
        types : list of strings or dicts
            List of lineType names or dicts for each section. If strings, 
            these names must match keys in the parent system lineTypes 
            dictionary or the subsystem's lineTypes dictionary. If dicts,
            these dicts are referred to for each lineType (by reference).
        connectors : list of dicts
            List of length nLines-1 with dicts of optional properties for
            any interior points (between sections).
        suspended : int
            Selector shared/suspended cases: 
            - 0 (default): end A is on the seabed,
            - 1: the assembly is suspended and end A is at another floating system,
            - 2: the assembly is suspended and assumed symmetric, end A is the midpoint.
        '''
        
        # some initialization steps.
        self.nLines = len(lengths)
        if len(connectors) == 0:
            connectors = [{}]*(self.nLines - 1)
        elif not len(connectors) == self.nLines - 1:
            raise Exception('Length of connectors must be nLines - 1')
        
        if not len(types)==self.nLines:
            raise Exception("The specified number of lengths and types is inconsistent.")
        
        # get cumulative sum of line lengths, starting from anchor segment
        Lcsum = np.cumsum(np.array(lengths))
        
        # set end A location depending on whether configuration is suspended/symmetrical
        if suspended==2:  # symmetrical suspended case
            rA = np.array([-0.5*self.span-self.rad_fair, 0, -1])  # shared line midpoint coordinates
            self.shared = True  # flag that it's being modeled as symmetric
        elif suspended==1:  # general suspended case
            rA = np.array([-self.span-self.rad_fair, 0, self.z_fair])  # other suspended end
        else:  # normal anchored line case
            rA = np.array([-self.span-self.rad_fair, 0, -self.depth])  # anchor coordinates
        rB = np.array([-self.rad_fair, 0, self.z_fair])     # fairlead coordinates

        self.rA = rA
        self.rB = rB

        if suspended==2:
            self.addPoint(0, rA, DOFs=[2]) # add shared line point, free only to move in z
        else:
            self.addPoint(-1, rA, DOFs=[0,2])                # add anchor point
        
        # Go through each line segment and add its upper point, add the line, and connect the line to the points
        for i in range(self.nLines):

            # find the specified lineType dict and save a reference to it
            if type(types[i]) == dict:  # if it's a dictionary, just point to it
                self.lineTypes[i] = types[i]
            # otherwise we're assuming it's a string of the lineType name
            elif types[i] in self.lineTypes:  # first look for the name in the subsystem
                self.lineTypes[i] = self.lineTypes[types[i]]
            elif self.sys: # otherwise look in the parent system, if there is one
                if types[i] in self.sys.lineTypes:  # first look for the name in the subsystem
                    self.lineTypes[i] = self.sys.lineTypes[types[i]]
                else:
                    raise Exception(f"Can't find lineType '{types[i]}' in the SubSystem or parent System.")
            else:
                raise Exception(f"Can't find lineType '{types[i]}' in the SubSystem.")
            
            # add the line segment using the reference to its lineType dict
            if nSegs is None:
                self.addLine(lengths[i], self.lineTypes[i])
            elif isinstance(nSegs, (int, float)):
                self.addLine(lengths[i], self.lineTypes[i], nSegs=nSegs)
            elif isinstance(nSegs, list):
                self.addLine(lengths[i], self.lineTypes[i], nSegs=nSegs[i])
            else:
                raise ValueError("Invalid type for nSegs. Expected None, a number, or a list.")

            # add the upper end point of the segment
            if i==self.nLines-1:                            # if this is the upper-most line
                self.addPoint(-1, rB, DOFs=[0,2])  # add the fairlead point (make it coupled)
                #self.bodyList[0].attachPoint(i+2, rB)       # attach the fairlead point to the body (two points already created)
            else:                                           # if this is an intermediate line
                m = connectors[i].get('m', 0)
                v = connectors[i].get('v', 0)
                # add the point, initializing linearly between anchor and fairlead/midpoint
                self.addPoint(0, rA + (rB-rA)*Lcsum[i]/Lcsum[-1], m=m, v=v, DOFs=[0,2])

            # attach the line to the points
            self.pointList[-2].attachLine(i+1, 0)       # attach end A of the line
            self.pointList[-1].attachLine(i+1, 1)       # attach end B of the line
        
    
    def setEndPosition(self, r, endB, sink=False):
        '''Sets either end position of the subsystem in the global/system
        reference frame. This is included mainly to mimic the Line method.

        Parameters
        ----------
        r : array
            x,y,z coorindate position vector of the line end [m].
        endB : boolean
            An indicator of whether the r array is at the end or beginning of the line
        sink : bool
            If true, and if there is a subsystem, the z position will be on the seabed.

        Raises
        ------
        LineError
            If the given endB value is not a 1 or 0
        '''
        if sink: # set z coordinate on seabed
            if self.sys:  # If there is a parent system, query it's seabed info
                z, _ = self.sys.getDepthFromBathymetry(r[0], r[1])
            else:
                z = -self.depth  # Otherwise use the subsystem's depth parameter
            
            r = np.array([r[0], r[1], -z])  # (making a copy of r to not overwrite it)
        
        # set end coordinates in global frame just like for a Line
        if endB == 1:
            self.rB = np.array(r, dtype=float)
        elif endB == 0:
            self.rA = np.array(r, dtype=float)
        else:
            raise LineError("setEndPosition: endB value has to be either 1 or 0")
        
        # Could optionally set end position here (though it will be set in staticSolve)
        
    
    def staticSolve(self, tol=0, profiles=0, maxIter=500):
        '''Solve internal equilibrium of the Subsystem and saves the forces
        and stiffnesses at the ends in the global reference frame. All the 
        This method mimics the behavior of the Line.staticSolve method and 
        the end coordinates rA and rB must be set beforehand. The solve 
        equilibrium happens in the local 2D plane. Values in this local 
        frame are also saved. 
        '''
        
        if tol==0:
            tol=self.eqtol
        
        # transform end positions to SubSystem internal coordinate system
        # inputs are self.rA, rB in global frame
        # outputs should be pointList[0] and [N] .r
        dr =  self.rB - self.rA
        LH = np.hypot(dr[0], dr[1])         # horizontal spacing of line ends
        LV = dr[2]                          # vertical rise from end A to B
        if self.shared:
            self.pointList[ 0].setPosition([  -self.span/2   , 0, self.rA[2]])
            self.pointList[-1].setPosition([ -self.span/2+LH, 0, self.rB[2]])
        else:
            self.pointList[ 0].setPosition([ -self.span   , 0, self.rA[2]])
            self.pointList[-1].setPosition([ -self.span+LH, 0, self.rB[2]])
        
        # transform coordinates and forces back into global frame
        if LH > 0:
            self.cos_th = dr[0]/LH          # cos of line heading
            self.sin_th = dr[1]/LH          # sin of line heading
            self.th = np.arctan2(dr[1],dr[0])    # heading (from A to B) [rad]
            self.R = rotationMatrix(0,0,self.th)  # rotation matrix about z that goes from +x direction to heading
            
        else:   # special case of vertical line: line heading is undefined - use zero as default
            self.cos_th = 1.0
            self.sin_th = 0.0
            self.th = 0
            self.R = np.eye(3)
        
        # update the bathymetry in case either end position has changed
        self.updateBathymetry()
            
        # get equilibrium
        self.solveEquilibrium(tol=tol, maxIter = maxIter)
        
        # get 2D stiffness matrices of end points
        K = self.getCoupledStiffnessA()
        
        # save end forces and stiffness matrices (first in local frame)
        self.fA_L = self.pointList[ 0].getForces(xyz=True) # force at end A
        self.fB_L = self.pointList[-1].getForces(xyz=True) # force at end B
        
        # Compute transverse (out-of-plane) stiffness term
        if LH < 0.01*abs(LV):  # if line is nearly vertical (note: this theshold is unverified)
            Kt = 0.5*(self.fA_L[2] - self.fB_L[2])/LV  # compute Kt based on vertical tension/span
        else:  # otherwise use the classic horizontal approach
            Kt = -self.fB_L[0]/LH  
        
        # expand to get 3D stiffness matrices
        '''
        R = np.eye(3)
        self.KA_L  = from2Dto3Drotated(K[:2,:2],  Kt, R.T)  # reaction at A due to motion of A
        self.KB_L  = from2Dto3Drotated(K[2:,2:],  Kt, R.T)  # reaction at B due to motion of B
        self.KBA_L = from2Dto3Drotated(K[2:,:2], -Kt, R.T)  # reaction at B due to motion of A
        '''
        self.KA_L = np.array([[K[0,0], 0 , K[0,1]],
                              [  0   , Kt,   0   ],
                              [K[1,0], 0 , K[1,1]]])
        
        # If symmetrical model, ignore midpoint stiffness and force
        if self.shared:  
            self.KB_L = np.array(self.KA_L)  # same stiffness as A
            self.fB_L = np.array([-self.fA_L[0], 0, self.fA_L[1]]) # mirror of fA
        else:
            self.KB_L = np.array([[K[2,2], 0 , K[2,3]],
                                  [  0   , Kt,   0   ],
                                  [K[3,2], 0 , K[3,3]]])
        
        self.KBA_L = -self.KB_L
        
        # Save tension magnitudes
        self.TA = np.linalg.norm(self.fA_L)  # tensions [N]
        self.TB = np.linalg.norm(self.fB_L)
        
        # Save rotated quantities for larger system in global reference frame
        self.fA = np.matmul(self.R, self.fA_L)
        self.fB = np.matmul(self.R, self.fB_L)
        
        self.KA  = np.matmul(np.matmul(self.R, self.KA_L ), self.R.T)
        self.KB  = np.matmul(np.matmul(self.R, self.KB_L ), self.R.T)
        self.KBA = np.matmul(np.matmul(self.R, self.KBA_L), self.R.T)
        
        #self.LBot = info["LBot"]  <<< this should be calculated considering all lines
        
        # ----- plot the profile if requested -----
        if profiles > 1:
            import matplotlib.pyplot as plt
            self.plot2d()
            plt.show()    
    
    
    def drawLine2d(self, Time, ax, color="k", endpoints=False, Xuvec=[1,0,0], Yuvec=[0,0,1], Xoff=0, Yoff=0, colortension=False, plotnodes=[], plotnodesline=[],label="",cmap='rainbow', alpha=1.0, linewidth = 1):
        '''wrapper to System.plot2d with some transformation applied'''
        
        for i, line in enumerate(self.lineList):
            if isinstance(label,list) and len(label)==len(self.lineList):
                lab = label[i]
            else:
                lab = label
            # color and width settings
            if color == 'self':
                colorplot = line.color  # attempt to allow custom colors
                lw = line.lw
            elif color == None:
                colorplot = [0.3, 0.3, 0.3]  # if no color, default to grey
                lw = 1
            else:
                colorplot = color
                lw = 1
            
            # get the Line's local coordinates
            Xs0, Ys0, Zs, tensions = line.getLineCoords(Time)
            
            # transform to global coordinates (Ys0 should be zero for now)
            Xs = self.rA[0] + (Xs0 + self.span)*self.cos_th - Ys0*self.sin_th
            Ys = self.rA[1] + (Xs0 + self.span)*self.sin_th + Ys0*self.cos_th
            
            # apply 3D to 2D transformation to provide desired viewing angle
            Xs2d = Xs*Xuvec[0] + Ys*Xuvec[1] + Zs*Xuvec[2] + Xoff
            Ys2d = Xs*Yuvec[0] + Ys*Yuvec[1] + Zs*Yuvec[2] + Yoff
            
            
            if colortension:    # if the mooring lines want to be plotted with colors based on node tensions
                maxT = np.max(tensions); minT = np.min(tensions)
                for i in range(len(Xs)-1):          # for each node in the line
                    color_ratio = ((tensions[i] + tensions[i+1])/2 - minT)/(maxT - minT)  # ratio of the node tension in relation to the max and min tension
                    cmap_obj = cm.get_cmap(cmap_tension)    # create a cmap object based on the desired colormap
                    rgba = cmap_obj(color_ratio)    # return the rbga values of the colormap of where the node tension is
                    ax.plot(Xs2d[i:i+2], Ys2d[i:i+2], color=rgba, zorder=100)
            else:
                ax.plot(Xs2d, Ys2d, color=colorplot, lw=lw, zorder=100, label=lab, alpha=alpha)
            
            if len(plotnodes) > 0:
                for i,node in enumerate(plotnodes):
                    if self.number==plotnodesline[i]:
                        ax.plot(Xs2d[node], Ys2d[node], 'o', color=colorplot, markersize=5)
            
            if endpoints == True:
                ax.scatter([Xs2d[0], Xs2d[-1]], [Ys2d[0], Ys2d[-1]], color = colorplot)


    def drawLine(self, Time, ax, color="k", endpoints=False, shadow=True, colortension=False, cmap_tension='rainbow'):
        '''wrapper to System.plot with some transformation applied'''
        
        for i, line in enumerate(self.lineList):
            
            # color and width settings
            if color == 'self':
                line.color = line.color  # attempt to allow custom colors
                line.lw = line.lw
            elif color == None:
                line.color = [0.3, 0.3, 0.3]  # if no color, default to grey
                line.lw = 1
            else:
                line.lw = 1
                line.color = color
            
            # get the Line's local coordinates
            Xs0, Ys0, Zs, tensions = line.getLineCoords(Time)
            
            # transform to global coordinates (Ys0 should be zero for now)
            Xs = self.rA[0] + (Xs0 + self.span)*self.cos_th - Ys0*self.sin_th
            Ys = self.rA[1] + (Xs0 + self.span)*self.sin_th + Ys0*self.cos_th
            
            if colortension:    # if the mooring lines want to be plotted with colors based on node tensions
                maxT = np.max(tensions); minT = np.min(tensions)
                for i in range(len(Xs)-1):          # for each node in the line
                    color_ratio = ((tensions[i] + tensions[i+1])/2 - minT)/(maxT - minT)  # ratio of the node tension in relation to the max and min tension
                    cmap_obj = cm.get_cmap(cmap_tension)    # create a cmap object based on the desired colormap
                    rgba = cmap_obj(color_ratio)    # return the rbga values of the colormap of where the node tension is
                    #linebit.append(ax.plot(Xs[i:i+2], Ys[i:i+2], Zs[i:i+2], color=rgba, zorder=100))
                    ax.plot(Xs[i:i+2], Ys[i:i+2], Zs[i:i+2], color=rgba, zorder=100)
            else:
                #linebit.append(ax.plot(Xs, Ys, Zs, color=color, lw=lw, zorder=100))
                ax.plot(Xs, Ys, Zs, color=line.color, lw=line.lw, zorder=100)
            
            if shadow:
                if self.seabedMod == 0:
                    Zs = np.zeros_like(Xs)-self.depth
                elif self.seabedMod == 1:
                    Zs = self.depth - self.xSlope*Xs - self.ySlope*Ys
                elif self.seabedMod == 2:
                    Zs = np.zeros(len(Xs))
                    for i in range(len(Xs)):
                        # Can get depths from the Subsystem's local grid
                        Zs[i] = self.getDepthFromBathymetry(Xs0[i], Ys0[i])
                        #Zs[i] = self.sys.getDepthFromBathymetry(Xs[i], Ys[i])

                ax.plot(Xs, Ys, Zs, color=[0.5, 0.5, 0.5, 0.2], lw=line.lw, zorder = 1.5) # draw shadow
            
            if endpoints == True:
                #linebit.append(ax.scatter([Xs[0], Xs[-1]], [Ys[0], Ys[-1]], [Zs[0], Zs[-1]], color = color))
                ax.scatter([Xs[0], Xs[-1]], [Ys[0], Ys[-1]], [Zs[0], Zs[-1]], color = line.color)
    
    
    def setOffset(self, offset, z=0):
        '''Moves end B of the Subsystem to represent an offset from the 
        undisplaced position of the endpoint. End A is set based on the 
        'span' (shouldn't change), and B is set based on offset and the
        rad_fair/z_fair setting. Optional argument z can be added for a z offset.
        '''
        
        self.offset = float(offset)
        
        # Use static EA values and unstretched lengths
        self.revertToStaticStiffness()

        # Ensure end A position and set end B position to offset values
        if self.shared:
            self.rA = np.array([-self.span/2-self.rad_fair, 0, self.rA[2]])
            self.rB = np.array([-self.rad_fair + offset/2, 0, self.z_fair+z]) 
            
        else:
            self.rA = np.array([-self.span-self.rad_fair, 0, self.rA[2]])
            self.rB = np.array([-self.rad_fair + offset, 0, self.z_fair+z]) 
            
        if hasattr(self, 'maxIter'):
            self.staticSolve(tol=self.eqtol, maxIter = self.maxIter)  # solve the subsystem
        else:
            self.staticSolve(tol=self.eqtol)
            
        # Store some values at this offset position that may be used later
        for i, line in enumerate(self.lineList):
            self.TeM[i,0] = np.linalg.norm(line.fA)
            self.TeM[i,1] = np.linalg.norm(line.fB)
               
        # save anchor forces at mean offset    
        self.anchorFxM = self.lineList[0].fA[0]
        self.anchorFzM = self.lineList[0].fA[2]
    
        self.TeD = np.copy(self.TeM)  # set the dynamic values as well, in case they get queried right away
        self.TeDmin = np.copy(self.TeM)  # set the dynamic values as well, in case they get queried right away
        
    
    def setDynamicOffset(self, offset, z=0):
        '''Moves end B of the Subsystem to represent a dynamic offset from
        the previous offset location. Uses dynamic stiffness values and also
        applies dynamic amplification factors (DAFs) on the difference from
        the mean tensions (which would have been calculated in getOffset).
        End A is set based on the 'span' (shouldn't change), 
        and B is set based on offset and the rad_fair/z_fair setting. 
        Optional argument z can be added for a z offset.
        '''
        
        self.offset = float(offset)
        
        if not self.dynamic_stiffness_activated: # if not already using them,
            System.activateDynamicStiffness(self)  # switch to dynamic EA values
        
        # adjust end B to the absolute offsets specified
        if self.shared:
            self.rB = np.array([-self.rad_fair + offset/2, 0, self.z_fair+z]) 
        else:
            self.rB = np.array([-self.rad_fair + offset, 0, self.z_fair+z]) 
        self.staticSolve(tol=self.eqtol)  # solve the subsystem
        
        # Store dynamic values at this offset position that may be used later
        for i, line in enumerate(self.lineList):
            self.TeD[i,:] = self.TeM[i,:] + self.DAFs[i]*( np.array([line.TA, line.TB]) - self.TeM[i,:] )
            self.TeDmin[i,:] = self.TeM[i,:] - self.DAFs[i]*( np.array([line.TA, line.TB]) - self.TeM[i,:] )
        
        # Store dynamic anchor forces
        self.anchorFxD = self.anchorFxM + self.DAFs[-1]*(self.lineList[0].fA[0] - self.anchorFxM)
        self.anchorFzD = self.anchorFzM + self.DAFs[-2]*(self.lineList[0].fA[2] - self.anchorFzM)
    
    def activateDynamicStiffness(self, display=0):
        '''Calls the dynamic stiffness method from System rather than from Line.'''
        System.activateDynamicStiffness(self, display=display)
    
    
    def revertToStaticStiffness(self):
        '''Calls the static stiffness method from System rather than from Line.'''
        System.revertToStaticStiffness(self)
    
    
    def getDynamicMatrices(self, omegas, S_zeta, r_dynamic, depth, kbot, cbot, seabed_tol=1e-4):
        '''Compute M,A,B,K matrices for the Subsystem. This calls 
        get_dynamic_matrices() for each Line in the Subsystem then combines
        the results. Note that this method overrides the Line method. Other
        Line methods used for dynamics can be used directly in Subsystem.
        '''
        self.nNodes = np.sum([line.nNodes for line in self.lineList]) - self.nLines + 1
        
        EA_segs = np.zeros(self.nNodes-1) # extensional stiffness of the segments
        n_dofs = 3*self.nNodes # number of dofs
        M = np.zeros([n_dofs,n_dofs], dtype='float')
        A = np.zeros([n_dofs,n_dofs], dtype='float')
        B = np.zeros([n_dofs,n_dofs], dtype='float')
        K = np.zeros([n_dofs,n_dofs], dtype='float')
        r_mean = np.zeros([self.nNodes,3], dtype='float')
        r_dynamic = np.ones((len(omegas),self.nNodes,3),dtype='float')*r_dynamic
        v_dynamic = 1j*omegas[:,None,None]*r_dynamic

        n = 0  # starting index of the next line's entries in the matrices
        
        for line in self.lineList:
            n1 = int(n/3) 
            n2 = n1 + line.nNodes

            # Filling matrices for line (dof n to dof 3xline_nodes+n)
            M_n,A_n,B_n,K_n,r_n,EA_segs_n = line.getDynamicMatrices(omegas, S_zeta,r_dynamic[:,n1:n2,:],depth,kbot,cbot,seabed_tol=seabed_tol)
            M[n:3*line.nNodes+n,n:3*line.nNodes+n] += M_n
            A[n:3*line.nNodes+n,n:3*line.nNodes+n] += A_n
            B[n:3*line.nNodes+n,n:3*line.nNodes+n] += B_n
            K[n:3*line.nNodes+n,n:3*line.nNodes+n] += K_n

            # Attachment point properties
            attachment = self.pointList[line.attached[-1]-1] # attachment point
            attachment_idx = n2 - 1 # last node index
            sigma_vp = np.sqrt(np.trapezoid(np.abs(v_dynamic[:,attachment_idx,:])**2*S_zeta[:,None],omegas,axis=0)) # standard deviations of the global components of the attachment point's velocity

            M[3*line.nNodes - 3:3*line.nNodes, 3*line.nNodes - 3:3*line.nNodes] += attachment.m*np.eye(3) 
            A[3*line.nNodes - 3:3*line.nNodes, 3*line.nNodes - 3:3*line.nNodes] += attachment.Ca* self.rho * attachment.v * np.eye(3)
            B[3*line.nNodes - 3:3*line.nNodes, 3*line.nNodes - 3:3*line.nNodes] += 0.5* self.rho * attachment.CdA * np.pi*(3*attachment.v/4/np.pi)**(2/3) \
                                                                                   * np.eye(3) * np.sqrt(8/np.pi) * np.diag(sigma_vp)

            # Static line properties
            r_mean[n1:n2,:] = r_n
            EA_segs[n1:n2-1] = EA_segs_n 

            n += 3*line.nNodes - 3 # next line starting node add the number of dofs of the current line minus 3 to get the last shared node
        
        return M,A,B,K,r_mean,EA_segs
    
    
    # ---- Extra convenience functions (subsystem should be in equilibrium) -----


    def getLayLength(self, iLine=0):
        '''Computes how much of a line is on the seabed.
           Finds the total LBot of the whole Mooring looking at all lines in the Mooring
        '''

        uplift_ang = np.degrees(np.arctan2(self.lineList[iLine].fA[2], self.lineList[iLine].fA[0]))

        LBot = sum([line.LBot for line in self.lineList if line.number <= self.nLines])

        if LBot == 0.0:  # if there is zero lay length, enter a negative value equal to the negative sign of the anchor inclination
            #LBot = -self.lineList[iLine].fA[2]/np.linalg.norm(self.lineList[iLine].fA) # this is to give the optimizer a gradient to help GET out of the constraint violation

            LBot = -uplift_ang
            # UPDATED: calculate the angle of uplift when LBot is negative to keep constraint 'continuous'
            # If input constraint value is negative, that's the highest angle (degrees) of vertical uplift allowed
        
        return LBot


    def getPointHeight(self, iPoint):
        '''Computes how high a point is off the seabed.
           Default is assuming point 1.
        '''

        if iPoint==0:   # if the input index of the point that needs to be constrained is the anchor point (0)
            Xs, Ys, Zs, Ts = self.lineList[iPoint].getLineCoords(0) # get the current coordinates of the nodes in the bottom line
            height_off_seabed = Zs[1] + self.depth              # the rope point that needs to be constrained is the first node of the line away from the anchor
        else:
            height_off_seabed = self.pointList[iPoint].r[2] + self.depth # measure the height of the rope-chain connecting point

        # if it's on the seabed, include some other factor to avoid a zero jacobian
        if height_off_seabed <= 0:
            fudge_factor = -self.lineList[iPoint].LBot       # add the negative of the contact length of the next line
        else:
            fudge_factor = 0.0

        if height_off_seabed < 0:
            print("Something is wrong")
            breakpoint()

        return height_off_seabed + fudge_factor

    
    def getHog(self, iLine):
        '''Compute the z elevation of the highest point along a line section.'''
        line = self.lineList[iLine]
        return max([line.z_extreme, line.rA[2], line.rB[2]])
    
    def getSag(self, iLine):
        '''Compute the z elevation of the lowest point along a line section.'''
        line = self.lineList[iLine]
        return min([line.z_extreme, line.rA[2], line.rB[2]])
    
    def getMinSag(self):
        '''Compute the z elevation of the lowest point of the whole subsystem.'''
        return min([ min([line.z_extreme, line.rA[2], line.rB[2]]) for line in self.lineList ])

    
    def getTen(self, iLine):
        '''Compute the end (maximum) tension for a specific line, including
        a dynamic amplification factor.'''
        line = self.lineList[iLine]
        '''
        dynamicTe = max([ self.Te0[iLine,0] + self.DAFs[iLine]*(np.linalg.norm(line.fA) - self.Te0[iLine,0]),
                          self.Te0[iLine,1] + self.DAFs[iLine]*(np.linalg.norm(line.fB) - self.Te0[iLine,1])])
        '''
        dynamicTe = max( self.TeD[iLine,:] )
        
        return dynamicTe 
    
    def getMinTen(self, iLine):
        '''Compute the end (maximum) tension for a specific line, including
        a dynamic amplification factor.'''
        line = self.lineList[iLine]

        dynamicTe = min( self.TeDmin[iLine,:] )
        
        return dynamicTe 
    
    
    def getTenSF(self, iLine, corrosion_mm=0):
        '''Compute MBL/tension for a specific line.'''
        return self.lineList[iLine].type['MBL'] / self.getTen(iLine) 
        

    def getMinTenSF(self, display=0):
        '''Compute minimum MBL/tension across all line sections.'''
        return min([self.getTenSF(i) for i in range(len(self.lineList))])
        
        
    def getCurv(self, iLine):
        '''Compute the maximum curvature (minimum bending radius) of a line
        section. In the future this could include a DAF.'''        
        
        # For now, this gets the max of the nodal curvatures for this line
        # which should already be calculated by calcCurvature (Is==iLine is
        # a mask). In future this could be replaced with an analytic calc.
        return max(self.Ks[self.Is==iLine]) 
    
    def getCurvSF(self, iLine):
        '''Compute minimum allowable bending radius / actual bend radius for
        a line/cable section.'''
        return 1 /( self.lineList[iLine].type['MBR'] * self.getCurv(iLine) )
    
    def getMinCurvSF(self, display=0):
        '''Compute minimum MBL/tension across all line sections.'''
        return min([self.getCurvSF(i) for i in range(len(self.lineList))])
        
    
    def getYawStiffness(self):
        '''Compute the contribution to platform yaw stiffness. Should already 
        be in equilibrium.'''
        
        tau0 = -self.fB[0]  # horizontal tension component [N]

        yaw_stiff = (tau0/l)*self.rad_fair**2 + tau0*self.rad_fair  # [N-m]
        
        return yaw_stiff


    def calcCurvature(self):
        '''Get the curvature [1/m] at each node of a Subsystem. Should already 
        be in equilibrium. Also populates nodal values for S, X, Y, Z, T, and K
        along the full length.'''
        
        n = self.nNodes
        
        self.Ss = np.zeros(n)  # arc length along line assembly (unstretched) [m]
        self.Xs = np.zeros(n)
        self.Ys = np.zeros(n)
        self.Zs = np.zeros(n)
        self.Ts = np.zeros(n)  # tensions [N]
        self.Ks = np.zeros(n)  # curvatures [1/m]
        self.Is = np.zeros(n, dtype=int)  # what line ID this node corresponds to (end A inclusive)
        
        ia = 0  # index of a Line's first node, as counted along the Subsystem
        La = 0  # length from end A of the Subsystem to end A of the Line
        
        for i, line in enumerate(self.lineList):
            ni = line.nNodes # number of nodes in this Line
            Ss = np.linspace(0, line.L, ni) + La
            Xs, Ys, Zs, Ts = line.getLineCoords(0)
            
            # paste in values for this line into the full array
            self.Ss[ia:ia+ni] = Ss
            self.Xs[ia:ia+ni] = Xs
            self.Ys[ia:ia+ni] = Ys
            self.Zs[ia:ia+ni] = Zs
            self.Ts[ia:ia+ni] = Ts
            self.Is[ia:ia+ni] = i
            
            # increment the starting index and length for the next line 
            ia += ni-1  # one point will be overwritten since lines share end points
            La += line.L
            
        
        #ds = np.diff(self.Ss)  # unstretched length of each segment
        ds = np.zeros(n-1)  # stretched length of each segment
        
        q = np.zeros([n-1, 3])  # unit vector of each segment
        
        
        # get length and direction of each segment
        for i in range(n-1):
            
            dr = np.array([self.Xs[i+1]-self.Xs[i], self.Ys[i+1]-self.Ys[i], self.Zs[i+1]-self.Zs[i]])
            
            ds[i] = np.linalg.norm(dr) # stretched length of each segment [m]
            
            q[i,:] = dr/ds[i]  # unit direction vector of the segment
            
            
        # now go through and get curvature at each node
        curvature = np.zeros(n)  # curvature at each node [1/m]
        for i in range(1,n-1):
            dl = (ds[i-1] + ds[i])/2  # average of two segment lengths
            
            curvature[i] = (2 / dl) * np.sqrt((1 - np.dot(q[i-1,:], q[i,:])) / 2)
        
        # assume the curvature of the end nodes is the same as the adjacent nodes
        curvature[ 0] = curvature[ 1]
        curvature[-1] = curvature[-2]
        
        # save values
        self.Ks = curvature

        return curvature
    
    
    def loadData(self, dirname, rootname, line_IDs):
        '''Load data from MoorDyn into the Subsystem lineList.
        
        dirname 
            folder where output files are located
        rootname
            root name of the OpenFAST primary file (.fst)
        line_IDs : list
            list of the line numbers (of MoorDyn) that correspond to the 
            lineList. These are the numbers of the line.out file to load.
        
        '''
        
        self.qs = 0  # indicate this Subsystem is reading in MoorDyn data
        
        for i, line in enumerate(self.lineList):
            line.loadData(dirname, rootname, line_IDs[i])






import numpy as np
from moorpy.helpers import transformPosition, rotationMatrix, rotatePosition, translateForce3to6DOF, getH, printVec



class Body():
    '''A class for any object in the mooring system that will have its own reference frame'''
    
    def __init__(self, mooringSys, num, type, r6, m=0, v=0, rCG=np.zeros(3),
                 AWP=0, rM=np.zeros(3), f6Ext=np.zeros(6), I=np.zeros(3),
                 CdA=np.zeros(3), Ca=np.zeros(3), DOFs=[0,1,2,3,4,5]):
        '''Initialize Body attributes

        Parameters
        ----------
        mooringSys : system object
            The system object that contains the body object
        num : int
            indentifier number
        type : int
            the body type: 0 free to move, 1 fixed, -1 coupled externally
        r6 : array
            6DOF position and orientation vector [m, rad]
        m : float, optional
            mass, centered at CG [kg]. The default is 0.
        v : float, optional
            volume, centered at reference point [m^3]. The default is 0.
        rCG : array, optional
            center of gravity position in body reference frame [m]. The default is np.zeros(3).
        AWP : float, optional
            waterplane area - used for hydrostatic heave stiffness if nonzero [m^2]. The default is 0.
        rM : float or array, optional
            coorindates or height of metacenter relative to body reference frame [m]. The default is np.zeros(3).
        f6Ext : array, optional
            applied external forces and moments vector in global orientation (not including weight/buoyancy) [N]. The default is np.zeros(6).
        I : array, optional
            Mass moment of inertia about 3 axes.
        CdA : array, optional
            Product of drag coefficient and frontal area in three directions [m^2].
        Ca : array, optional
            Added mass coefficient in three directions.
        DOFs: list, optional
            list of the DOFs for this body (0=surge,1=sway,...5=yaw). Any not 
            listed will be held fixed. E.g. [0,1,5] for 2D horizontal motion.
        
        Returns
        -------
        None.

        '''
    
        self.sys    = mooringSys       # store a reference to the overall mooring system (instance of System class)
        
        self.number = num
        self.type   = type                          # 0 free to move, or -1 coupled externally
        self.r6     = np.array(r6, dtype=float)     # 6DOF position and orientation vector [m, rad]
        self.m      = m                             # mass, centered at CG [kg]
        self.v      = v                             # volume, assumed centered at reference point [m^3]
        self.rCG    = np.array(rCG, dtype=float)    # center of gravity position in body reference frame [m]
        self.AWP    = AWP                           # waterplane area - used for hydrostatic heave stiffness if nonzero [m^2]
        if np.isscalar(rM):
            self.rM = np.array([0,0,rM], dtype=float) # coordinates of body metacenter relative to body reference frame [m]
        else:
            self.rM = np.array(rM, dtype=float)       

        # >>> should streamline the below <<<
        if np.isscalar(I):
            self.I = np.array([I,I,I], dtype=float)
        else:
            self.I = np.array(I, dtype=float)    

        if np.isscalar(CdA):
            self.CdA = np.array([CdA,CdA,CdA], dtype=float)
        else:
            self.CdA = np.array(CdA, dtype=float)    
            
        if np.isscalar(Ca):
            self.Ca = np.array([Ca,Ca,Ca], dtype=float) 
        else:
            self.Ca = np.array(Ca, dtype=float)                
                
        self.f6Ext  = np.array(f6Ext, dtype=float)    # for adding external forces and moments in global orientation (not including weight/buoyancy)
        
        self.DOFs = DOFs
        self.nDOF = len(DOFs)
        
        self.attachedP   = []          # ID numbers of any Points attached to the Body
        self.rPointRel   = []          # coordinates of each attached Point relative to the Body reference frame
        
        self.attachedR   = []          # ID numbers of any Rods attached to the Body (not yet implemented)
        self.r6RodRel   = []           # coordinates and unit vector of each attached Rod relative to the Body reference frame
        
        self.R = np.eye(3)             # body orientation rotation matrix
        #print("Created Body "+str(self.number))
        
    
    def attachPoint(self, pointID, rAttach):
        '''Adds a Point to the Body, at the specified relative position on the body.
        
        Parameters
        ----------
        pointID : int
            The identifier ID number of a point
        rAttach : array
            The position of the point relative to the body's frame [m]
        '''
    
        self.attachedP.append(pointID)
        self.rPointRel.append(np.array(rAttach))
        
        if self.sys.display > 1:
            print("attached Point "+str(pointID)+" to Body "+str(self.number))
    
    """
    def dettachPoint(self, pointID):
        '''Removes a Point from the Body.
        
        Parameters
        ----------
        pointID : int
            The identifier ID number of a point
        rAttach : array
            The position of the point relative to the body's frame [m]
        '''
    
        find pointID index in self.attachedP
        delete entry from self.attachedP and rPointRel
        
        set point type to free
    """
    
    def attachRod(self, rodID, endCoords):
        '''Adds a Point to the Body, at the specified relative position on the body.
        
        Parameters
        ----------
        rodID : int
            The identifier ID number of a point
        endCoords : array
            The position of the Rods two ends relative to the body reference frame [m]

        Returns
        -------
        None.

        '''
    
        k = (endCoords[3:]-endCoords[:3])/np.linalg.norm(endCoords[3:]-endCoords[:3])
    
        self.attachedR.append(rodID)
        self.r6RodRel.append(np.hstack([ endCoords[:3], k]))
        
        print("attached Rod "+str(rodID)+" to Body "+str(self.number))
        
    
    def setPosition(self, r6):
        '''Sets the position of the Body, along with that of any dependent objects.

        Parameters
        ----------
        r6 : array
            6DOF position and orientation vector of the body [m, rad]

        Raises
        ------
        ValueError
            If the length of the input r6 array is not of length 6

        Returns
        -------
        None.

        '''
        
        # update the position/orientation of the body
        if len(r6)==6:
            self.r6 = np.array(r6, dtype=float)  # update the position of the Body itself
        elif len(r6) == self.nDOF:
            self.r6[self.DOFs] = r6  # mapping to use only some DOFs
        else:
            raise ValueError(f"Body setPosition method requires an argument of size 6 or nDOF, but size {len(r6):d} was provided")
        
        self.R = rotationMatrix(self.r6[3], self.r6[4], self.r6[5])   # update body rotation matrix
        
        # update the position of any attached Points
        for PointID,rPointRel in zip(self.attachedP,self.rPointRel):
            rPoint = np.matmul(self.R, rPointRel) + self.r6[:3]  # rPoint = transformPosition(rPointRel, r6)            
            self.sys.pointList[PointID-1].setPosition(rPoint)
            
        # update the position of any attached Rods        
        for rodID,r6Rel in zip(self.attachedR,self.r6RodRel):        
            rA = np.matmul(self.R, r6Rel[:3]) + self.r6[:3] 
            k = np.matmul(self.R, r6Rel[3:])
            self.sys.rodList[rodID-1].rA = rA
            self.sys.rodList[rodID-1].rB = rA + k*self.sys.rodList[rodID-1].L
    
    
    def getForces(self, lines_only=False, all_DOFs=False):
        '''Sums the forces and moments on the Body, including its own plus those from any attached objects.
        Forces and moments are aligned with global x/y/z directions but are relative 
        to the body's local reference point.

        Parameters
        ----------
        lines_only : boolean, optional
            If true, the Body's contribution to the forces is ignored.
        all_DOFs : boolean, optional
            True: return all forces/moments; False: only those in DOFs list.

        Returns
        -------
        f6 : array
            The 6DOF forces and moments applied to the body in its current position [N, Nm]
        '''
    
        f6 = np.zeros(6)
    
        # TODO: could save time in below by storing the body's rotation matrix when it's position is set rather than 
        #       recalculating it in each of the following function calls.
                
        if lines_only==False:
        
            # add weight, which may result in moments as well as a force
            rCG_rotated = rotatePosition(self.rCG, self.r6[3:])                     # relative position of CG about body ref point in unrotated reference frame  
            f6 += translateForce3to6DOF(rCG_rotated, np.array([0,0, -self.m*self.sys.g]))    # add to net forces/moments
        
            # add buoyancy force and moments if applicable (this can include hydrostatic restoring moments 
            # if rM is considered the metacenter location rather than the center of buoyancy)
            rM_rotated = rotatePosition(self.rM, self.r6[3:])                       # relative position of metacenter about body ref point in unrotated reference frame  
            f6 += translateForce3to6DOF(rM_rotated, np.array([0,0, self.sys.rho*self.sys.g*self.v]))  # add to net forces/moments
        
            # add hydrostatic heave stiffness (if AWP is nonzero)
            f6[2] -= self.sys.rho*self.sys.g*self.AWP*self.r6[2]
            
            # add any externally applied forces/moments (in global orientation)
            f6 += self.f6Ext
	
        # add forces from any attached Points (and their attached lines)
        for PointID,rPointRel in zip(self.attachedP,self.rPointRel):
        
            fPoint = self.sys.pointList[PointID-1].getForces(lines_only=lines_only, xyz=True) # get net force on attached Point
            rPoint_rotated = rotatePosition(rPointRel, self.r6[3:])                 # relative position of Point about body ref point in unrotated reference frame  
            f6 += translateForce3to6DOF(rPoint_rotated, fPoint)                     # add net force and moment resulting from its position to the Body
            
            
        # All forces and moments on the body should now be summed, and are in global/unrotated orientations.
        '''
        # For application to the body DOFs, convert the moments to be about the body's local/rotated x/y/z axes <<< do we want this in all cases? 
        rotMat = rotationMatrix(*self.r6[3:])                                       # get rotation matrix for body
        moment_about_body_ref = np.matmul(rotMat.T, f6[3:])                         # transform moments so that they are about the body's local/rotated axes
        f6[3:] = moment_about_body_ref                                              # use these moments
        '''
        
        if all_DOFs:
            return f6
        else:
            return f6[self.DOFs]

    
    def getStiffness(self, X = [], tol=0.0001, dx = 0.1, all_DOFs=False):
        '''Gets the stiffness matrix of a Body due only to mooring lines with all other objects free to equilibriate.
        The rotational indices of the stiffness matrix correspond to the global x/y/z directions.
        
        Parameters
        ----------
        X1 : array
            The position vector (6DOF) of the main axes of the Body at which the stiffness matrix is to be calculated.
        dx : float, optional
            The change in displacement to be used for calculating the change in force. The default is 0.01.
        all_DOFs : boolean, optional
            True: return all forces/moments; False: only those in DOFs list.

        Returns
        -------
        K : matrix
            The stiffness matrix of the body at the given position X1.
            
        '''
        
        #print("Getting Body "+str(self.number)+" stiffness matrix...")
        
        if len(X) == 6:
            X1 = np.array(X)
        elif len(X)==0:
            X1 = self.r6
        else:
            raise ValueError('Body.getStiffness expects the optional X parameter to be size 6')
        
        # set this Body's type to fixed so mooring system equilibrium response to its displacements can be found
        type0 = self.type                         # store original type to restore later
        self.type = 1                             # set type to 1 (not free) so that it won't be adjusted when finding equilibrium
        
        # ensure this Body is positioned at the desired linearization point
        self.setPosition(X1)                      # set position to linearization point
        self.sys.solveEquilibrium3(tol=tol)       # find equilibrium of mooring system given this Body in current position
        f6 = self.getForces(lines_only=True)      # get the net 6DOF forces/moments from any attached lines 
        
        # Build a stiffness matrix by perturbing each DOF in turn
        K = np.zeros([6,6])
        
        for i in range(len(K)):
            X2 = X1 + np.insert(np.zeros(5),i,dx) # calculate perturbed Body position by adding dx to DOF in question            
            self.setPosition(X2)                  # perturb this Body's position
            self.sys.solveEquilibrium3(tol=tol)   # find equilibrium of mooring system given this Body's new position
            f6_2 =self.getForces(lines_only=True) # get the net 6DOF forces/moments from any attached lines 
            
            K[:,i] = -(f6_2-f6)/dx                # get stiffness in this DOF via finite difference and add to matrix column
            
        # ----------------- restore the system back to previous positions ------------------
        self.setPosition(X1)                      # set position to linearization point
        self.sys.solveEquilibrium3(tol=tol)       # find equilibrium of mooring system given this Body in current position
        self.type = type0                         # restore the Body's type to its original value
        
        # Return stiffness matrix
        if all_DOFs:
            return K
        else:  # only return rows/columns of active DOFs
            return K[:,self.DOFs][self.DOFs,:]
        
    
    def getStiffnessA(self, lines_only=False, all_DOFs=False):
        '''Gets the analytical stiffness matrix of the Body with other objects fixed.

        Parameters
        ----------
        lines_only : boolean, optional
            If true, the Body's contribution to its stiffness is ignored.
        all_DOFs : boolean, optional
            True: return all forces/moments; False: only those in DOFs list.

        Returns
        -------
        K : matrix
            6x6 analytic stiffness matrix.
        '''
                
        K = np.zeros([6,6])
        
        
        # stiffness contributions from attached points (and any of their attached lines)
        for PointID,rPointRel in zip(self.attachedP,self.rPointRel):
            
            r = rotatePosition(rPointRel, self.r6[3:])          # relative position of Point about body ref point in unrotated reference frame  
            f3 = self.sys.pointList[PointID-1].getForces(xyz=True)      # total force on point (for additional rotational stiffness term due to change in moment arm)
            K3 = self.sys.pointList[PointID-1].getStiffnessA(xyz=True)  # local 3D stiffness matrix of the point
            
            # following are from functions translateMatrix3to6
            H = getH(r)
            K[:3,:3] += K3
            K[:3,3:] += np.matmul(K3, H)                        # only add up one off-diagonal sub-matrix for now, then we'll mirror at the end
            K[3:,3:] += -np.matmul(getH(f3), H) - np.matmul(H, np.matmul(K3,H))   # updated 2023-05-02
   
        K[3:,:3] = K[:3,3:].T                                   # copy over other off-diagonal sub-matrix
        
        
        # body's own stiffness components
        if lines_only == False:
        
            # rotational stiffness effect of weight
            rCG_rotated = rotatePosition(self.rCG, self.r6[3:]) # relative position of CG about body ref point in unrotated reference frame  
            Kw = -np.matmul( getH([0,0, -self.m*self.sys.g]) , getH(rCG_rotated) )
            
            # rotational stiffness effect of buoyancy at metacenter
            rM_rotated = rotatePosition(self.rM, self.r6[3:])   # relative position of metacenter about body ref point in unrotated reference frame  
            Kb = -np.matmul( getH([0,0, self.sys.rho*self.sys.g*self.v]) , getH(rM_rotated) )
           
            # hydrostatic heave stiffness (if AWP is nonzero)
            Kwp = self.sys.rho*self.sys.g*self.AWP
            
            K[3:,3:] += Kw + Kb
            K[2 ,2 ] += Kwp
            
        # Return stiffness matrix
        if all_DOFs:
            return K
        else:  # only return rows/columns of active DOFs
            return K[:,self.DOFs][self.DOFs,:]


    def getDynamicMatrices(self, lines_only=False, all_DOFs=False):
        '''Gets the dynamic matrices of the Body with other objects fixed
        using a lumped mass approach.

        Parameters
        ----------
        lines_only : boolean, optional
            If true, the Body's contribution to its stiffness is ignored.
        all_DOFs : boolean, optional
            True: return all forces/moments; False: only those in DOFs list.

        Returns
        -------
        M, A, B, K: 6x6 inertia, added mass, damping, and stiffness matrices.
        '''
                
        M, A, B, K = (np.zeros([6,6]) for _ in range(4))
        
        # Contributions from attached points (and any of their attached lines)
        for PointID,rPointRel in zip(self.attachedP,self.rPointRel):            
            r = rotatePosition(rPointRel, self.r6[3:])          # relative position of Point about body ref point in unrotated reference frame  
            f3 = self.sys.pointList[PointID-1].getForces()      # total force on point (for additional rotational stiffness term due to change in moment arm)

            M3, A3, B3, K3 = self.sys.pointList[PointID-1].getDynamicMatrices()  # local 3D dynamic matrices of the point
            
            # following are from functions translateMatrix3to6            
            H = getH(r)
            K[:3,:3] += K3
            K[:3,3:] += np.matmul(K3, H)                        # only add up one off-diagonal sub-matrix for now, then we'll mirror at the end
            K[3:,3:] += -np.matmul(getH(f3), H) - np.matmul(H, np.matmul(K3,H))   # updated 2023-05-02

            # Inertia, added mass, and damping matrices are similar. Except for the lower right submatrix, which does not have the first term because 
            # the moment arm does not explictly change with the velocity or acceleration of the body.
            M[:3,:3] += M3
            M[:3,3:] += np.matmul(M3, H)
            M[3:,3:] += - np.matmul(H, np.matmul(M3,H))

            A[:3,:3] += A3
            A[:3,3:] += np.matmul(A3, H)
            A[3:,3:] += - np.matmul(H, np.matmul(A3,H))

            B[:3,:3] += B3
            B[:3,3:] += np.matmul(B3, H)
            B[3:,3:] += - np.matmul(H, np.matmul(B3,H))            
   
        K[3:,:3] = K[:3,3:].T                                   # copy over other off-diagonal sub-matrix
        M[3:,:3] = M[:3,3:].T
        A[3:,:3] = A[:3,3:].T
        B[3:,:3] = B[:3,3:].T        
        
        # body's own stiffness components
        if lines_only == False:
        
            # rotational stiffness effect of weight
            rCG_rotated = rotatePosition(self.rCG, self.r6[3:]) # relative position of CG about body ref point in unrotated reference frame  
            Kw = -np.matmul( getH([0,0, -self.m*self.sys.g]) , getH(rCG_rotated) )
            
            # rotational stiffness effect of buoyancy at metacenter
            rM_rotated = rotatePosition(self.rM, self.r6[3:])   # relative position of metacenter about body ref point in unrotated reference frame  
            Kb = -np.matmul( getH([0,0, self.sys.rho*self.sys.g*self.v]) , getH(rM_rotated) )
           
            # hydrostatic heave stiffness (if AWP is nonzero)
            Kwp = self.sys.rho*self.sys.g*self.AWP
            
            K[3:,3:] += Kw + Kb
            K[2 ,2 ] += Kwp
            
        # Return stiffness matrix
        if all_DOFs:
            return M, A, B, K
        else:  # only return rows/columns of active DOFs
            return M[:,self.DOFs][self.DOFs,:], A[:,self.DOFs][self.DOFs,:], B[:,self.DOFs][self.DOFs,:], K[:,self.DOFs][self.DOFs,:]

    
    def draw(self, ax):
        '''Draws the reference axis of the body

        Parameters
        ----------
        ax : axes
            matplotlib.pyplot axes to be used for drawing and plotting.

        Returns
        -------
        linebit : list
            a list to hold plotted lines of the body's frame axes.

        '''
        
        linebit = []  # make empty list to hold plotted lines, however many there are
    
        rx = transformPosition(np.array([5,0,0]),self.r6)
        ry = transformPosition(np.array([0,5,0]),self.r6)
        rz = transformPosition(np.array([0,0,5]),self.r6)
        
        linebit.append(ax.plot([self.r6[0], rx[0]], [self.r6[1], rx[1]], [self.r6[2], rx[2]], color='r'))
        linebit.append(ax.plot([self.r6[0], ry[0]], [self.r6[1], ry[1]], [self.r6[2], ry[2]], color='g'))
        linebit.append(ax.plot([self.r6[0], rz[0]], [self.r6[1], rz[1]], [self.r6[2], rz[2]], color='b'))
        
        self.linebit = linebit
            
        return linebit
    
    
    def redraw(self):
        '''Redraws the reference axis of the body

        Returns
        -------
        linebit : list
            a list to hold redrawn lines of the body's frame axes.

        '''
    
        linebit = self.linebit
    
        rx = transformPosition(np.array([5,0,0]),self.r6)
        ry = transformPosition(np.array([0,5,0]),self.r6)
        rz = transformPosition(np.array([0,0,5]),self.r6)
        
        linebit[0][0].set_data_3d([self.r6[0], rx[0]], [self.r6[1], rx[1]], [self.r6[2], rx[2]])
        linebit[1][0].set_data_3d([self.r6[0], ry[0]], [self.r6[1], ry[1]], [self.r6[2], ry[2]])
        linebit[2][0].set_data_3d([self.r6[0], rz[0]], [self.r6[1], rz[1]], [self.r6[2], rz[2]])
        '''
        linebit[0][0].set_data([self.r6[0], rx[0]], [self.r6[1], rx[1]])
        linebit[0][0].set_3d_properties([self.r6[2], rx[2]])
        linebit[1][0].set_data([self.r6[0], ry[0]], [self.r6[1], ry[1]]) 
        linebit[1][0].set_3d_properties([self.r6[2], ry[2]])
        linebit[2][0].set_data([self.r6[0], rz[0]], [self.r6[1], rz[1]]) 
        linebit[2][0].set_3d_properties([self.r6[2], rz[2]])
        '''
        return linebit
    



import numpy as np
import matplotlib.pyplot as plt
from matplotlib import cm
from moorpy.helpers import CatenaryError



def nonlinear(XF, ZF, L, Str, Ten, W, nNodes=20, plots=0):
    '''
    nonlinear mooring line solver.  The nonlinear module is assumed to be off of the seafloor, and have no bending.  The nonlinear module is called by
    putting a tension-strain file in the place of linestiffness (EA).  Unlike the catenary solver no iterative method is needed, but this function is heavily 
    borrows from the catenary function.
    
    Parameters
    ----------
    XF : float
        Horizontal distance from end 1 to end 2 [m]
    ZF : float
        Vertical distance from end 1 to end 2 [m] (positive up)
    L  : float
        Unstretched length of line [m]
    Str : dictionary
        vector of strains from the tension-strain input file [-]
    Ten : dictionary 
        vector of  of tensions from the tension-strain input file [N]    
    W  : float
        Weight of line in fluid per unit length [N/m]   
    nNodes : int, optional
        Number of nodes to describe the line
    plots  : int, optional
        1: plot output, 0: don't
    
    
    Returns
    -------
    : tuple
        (end 1 horizontal tension, end 1 vertical tension, end 2 horizontal tension, end 2 vertical tension, info dictionary) [N] (positive up)
    
    '''
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # The following is originally from the catenary.py function but applies to this one aswell
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    # make info dict to contain any additional outputs
    info = dict(error=False)
    
    # make some arrays if needed for plotting each node
    if plots > 0:
        s = np.linspace(0,L,nNodes)   #  Unstretched arc distance along line from anchor to each node where the line position and tension can be output (meters)
        Xs= np.zeros(nNodes)          #  Horizontal locations of each line node relative to the anchor (meters)
        Zs= np.zeros(nNodes)          #  Vertical   locations of each line node relative to the anchor (meters)
        Te= np.zeros(nNodes)          #  Effective line tensions at each node (N)

    
    # flip line in the solver if it is buoyant
    if W < 0:
        W = -W
        ZF = -ZF
        CB = -10000.   # <<< TODO: set this to the distance to sea surface <<<
        flipFlag = True
    else:
        flipFlag = False
    
    # reverse line in the solver if end A is above end B
    if ZF < 0:    
        ZF = -ZF
        reverseFlag = True   
    else:
        reverseFlag = False
    
    # ensure the input variables are realistic
    if XF < 0.0:
        raise CatenaryError("XF is negative!")
    if L <= 0.0:
        breakpoint()
        raise CatenaryError("L is zero or negative!")

    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    #    New Code
    #~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    
    #Determine the stretched length of the nonlinear element based on the horizontal and vertical fairlead excursions
    str_L = np.sqrt(XF*XF+ZF*ZF)
    
    #Based on the stretched length and the horizontal fairlead excursions define cos/sin and products for transformation matrix
    #Define cosine and sine
    c = XF/str_L
    s = ZF/str_L
    #Define all the products of sine and cosine
    c2 = c*c
    s2 = s*s
    cs = c*s
  
    #Determine the strain in the segment with stretched length and original length
    str_seg = (str_L/L - 1)
    
    #Make the input dictionarys for str and ten a vector of floats  
    Str = [float(x) for x in Str]
    Ten = [float(x) for x in Ten]
    
    #Strain in the line cant be less than zero (cant push a rope so to speak)
    if str_seg < 0:
        str_seg = 0 

    ten_seg = np.interp(str_seg,Str,Ten)
    #small change is strain for a finite difference
    d_str = 1e-5
    #finite difference to take derivative of ten-strain curve (dT/dstr = EA)
    EA_est = ((np.interp(str_seg+d_str,Str,Ten)-np.interp(str_seg,Str,Ten))/d_str)

    #Breakdown the horizontal components of the tension
    FxA = ten_seg*c
    FxB = -FxA

    #Breakdown the vertical components of the tension and subtract the weight of the nonlinear element
    FzA = ten_seg*s - 0.5*W*L
    FzB = -ten_seg*s - 0.5*W*L
    
    #Tension at both ends of the line
    TA = np.sqrt(FxA*FxA + FzA*FzA)
    TB = np.sqrt(FxB*FxB + FzB*FzB)
    
    #Estimate Stiffness Numerically (Potentially find something better)
    #could also be hard to do because the input tension strain files are arbitrary

    

    #Stiffness matrcies for the ends of the line (In this case since we already assume the element can only deform axially
    # this is essentially just a truss element with the nonlinear stiffness that we calculated
    #Ka = (EA_est/L)*np.array([[c2, cs],[cs, s2]])    
    #Kb = (EA_est/L)*np.array([[c2, cs],[cs, s2]]) 
    #Kab = -(EA_est/L)*np.array([[c2, cs],[cs, s2]]) 
    
    Ka = np.array([[c2*(EA_est/str_L)+s2*ten_seg/(L*(1+str_seg)), cs*(EA_est/str_L)-cs*ten_seg/(L*(1+str_seg))],[ cs*(EA_est/str_L)-cs*ten_seg/(L*(1+str_seg)), s2*(EA_est/str_L)+c2*ten_seg/(L*(1+str_seg))]])    
    Kb = Ka
    Kab = -Ka 
    
   
    #Assign values to info
    #Fairlead forces
    info["HF"] = -FxB
    info["VF"] = -FzB
    info["HA"] = FxA
    info["VA"] = FzA
    #Line stiffnesses
    info["stiffnessA"] = Ka
    info["stiffnessB"] = Kb
    info["stiffnessAB"] = Kab
    #Length on the bottom is zero as we assumed earlier
    info["LBot"] = 0
    #self.info = info
    #For plotting (assumed straight line so no issue and tension varies linearly
    info["X" ] = np.linspace(0,XF,nNodes)
    info["Z" ] = np.linspace(0,ZF,nNodes)
    info["s" ] = np.linspace(0,str_L,nNodes)
    info["Te"] = np.linspace(TA,TB,nNodes) 
    
    print(info["stiffnessA"])
    print(info["stiffnessB"])
    print(info["stiffnessAB"])


    return (FxA, FzA, FxB, FzB, info)     

 


# tests MoorPy Subsystem functionality and results

import pytest

from numpy.testing import assert_allclose

import numpy as np
import moorpy as mp
#from moorpy.MoorProps import getLineProps
from moorpy.helpers import getLineProps

import matplotlib.pyplot as plt


"""
def test_tensions_swap():
    '''Compares two equivalent catenary mooring lines that are defined in opposite directions.'''
       
    ms = mp.System(depth=60)

    #ms.lineTypes['chain'] = getLineProps(120, name='chain')    # add a line type
    ms.setLineType(120, 'chain', name='chain')    # add a line type
    
    ms.addPoint(1, [  0, 0, -60])
    ms.addPoint(1, [100, 10, -30])

    # line sloping up from A to B, and another in the opposite order
    ms.addLine(120, 'chain', pointA=1, pointB=2)
    ms.addLine(120, 'chain', pointA=2, pointB=1)

    ms.initialize()
    
    # compare tensions
    assert_allclose(np.hstack([ms.lineList[0].fA, ms.lineList[0].fB]),
                    np.hstack([ms.lineList[1].fB, ms.lineList[1].fA]), rtol=0, atol=10.0, verbose=True)
"""

# TO DO: make some test functions like the above one to test out some simple Subsystems for code bugs
# One key test is a 2-line subsystem that is fully slack with the middle point on the seabed, as in the below code.


"""

import numpy as np
import matplotlib.pyplot as plt
import moorpy as mp
from moorpy.MoorProps import getLineProps
from moorpy.subsystem import Subsystem


# ----- choose some system geometry parameters -----

depth     = 200                             # water depth [m]
angles    = np.radians([60, 300])      # line headings list [rad]
rAnchor   = 600                            # anchor radius/spacing [m]
zFair     = -21                             # fairlead z elevation [m]
rFair     = 20                              # fairlead radius [m]
lineLength= 650                            # line unstretched length [m]
typeName  = "chain1"                        # identifier string for the line type


# ===== First a simple Subsystem by itself =====

# create a subsystem
ss = Subsystem(depth=depth, spacing=rAnchor, rBfair=[10,0,-20])

# set up the line types
ss.setLineType(180, 'chain', name='one')
ss.setLineType( 50, 'chain', name='two')

# set up the lines and points and stuff
lengths = [350, 400]
types = ['one', 'two']
ss.makeGeneric(lengths, types)

# plotting examples
ss.setEndPosition([0  ,-40,-200], endB=0)
ss.setEndPosition([300,400,  -10], endB=1)
ss.staticSolve()
# 3D plot in the Subsystem's local frame
fig, ax = ss.plot()
# 2D view of the same
ss.plot2d()
# Line-like plot (in global reference frame)
fig = plt.figure()
ax = plt.axes(projection='3d')
ss.drawLine(0, ax, color='r')


# ===== Now a Subsystem in a larger System with a floating body =====

# Create new MoorPy System and set its depth
ms = mp.System(depth=depth)

# add a line type
ms.setLineType(dnommm=120, material='chain', name=typeName)  # this would be 120 mm chain

# Add a free, body at [0,0,0] to the system (including some properties to make it hydrostatically stiff)
ms.addBody(0, np.zeros(6), m=1e6, v=1e3, rM=100, AWP=1e3)

# For each line heading, set the anchor point, the fairlead point, and the line itself
for i, angle in enumerate(angles):

    # create end Points for the line
    ms.addPoint(1, [rAnchor*np.cos(angle), rAnchor*np.sin(angle), -depth])   # create anchor point (type 0, fixed)
    ms.addPoint(1, [  rFair*np.cos(angle),   rFair*np.sin(angle),  zFair])   # create fairlead point (type 0, fixed)
    
    # attach the fairlead Point to the Body (so it's fixed to the Body rather than the ground)
    ms.bodyList[0].attachPoint(2*i+2, [rFair*np.cos(angle), rFair*np.sin(angle), zFair]) 

    # add a Line going between the anchor and fairlead Points
    ms.addLine(lineLength, typeName, pointA=2*i+1, pointB=2*i+2)

# ----- Now add a SubSystem line! -----
ss = Subsystem(mooringSys=ms, depth=depth, spacing=rAnchor, rBfair=[10,0,-20])

# set up the line types
ms.setLineType(180, 'chain', name='one')
ms.setLineType( 50, 'chain', name='two')

# set up the lines and points and stuff
ls = [350, 300]
ts = ['one', 'two']
ss.makeGeneric(lengths=ls, types=ts)

# add points that the subSystem will attach to...
ms.addPoint(1, [-rAnchor, 100, -depth])   # Point 5 - create anchor point (type 0, fixed)
ms.addPoint(1, [ -rFair ,   0,  zFair])   # Point 6 - create fairlead point (type 0, fixed)
ms.bodyList[0].attachPoint(6, [-rFair, 0, zFair])  # attach the fairlead Point to the Body 

# string the Subsystem between the two points!
ms.lineList.append(ss)  # add the SubSystem to the System's lineList
ss.number = 3
ms.pointList[4].attachLine(3, 0)  # attach it to the respective points
ms.pointList[5].attachLine(3, 1)  # attach it to the respective points


# ----- run the model to demonstrate -----

ms.initialize()                                             # make sure everything's connected

ms.solveEquilibrium()                                       # equilibrate
fig, ax = ms.plot()                                         # plot the system in original configuration
#ms.unload("sample_from_manual.txt")                         # export to MD input file

ms.bodyList[0].f6Ext = np.array([3e6, 0, 0, 0, 0, 0])       # apply an external force on the body 
ms.solveEquilibrium3()                                      # equilibrate
fig, ax = ms.plot(ax=ax, color='red')                       # plot the system in displaced configuration (on the same plot, in red)

print(f"Body offset position is {ms.bodyList[0].r6}")

plt.show()
"""
